import pytest

from arthurai.common.exceptions import ArthurUserError
from arthurai.core.models import ArthurModel
from arthurai.core.alerts import AlertRule, AlertRuleBound, AlertRuleSeverity, validate_parameters_for_alert
from arthurai.client.http.requests import HTTPClient
from tests import MockResponse

alert_rule_object = AlertRule(
    bound=AlertRuleBound.Upper,
    threshold=0.5,
    metric_id="4e18df3d-307d-4b3f-aba4-a434ca893c8a",
    severity=AlertRuleSeverity.Warning,
    metric_parameters={
        "classifier_threshold": 0.5,
        "predicted_property": "prediction"
    },
    filters=[
        {
            "property": "AGE",
            "comparator": "eq",
            "value": 22
        }
    ]
)

alert_rule_dict = {
    "bound": "upper",
    "threshold": 0.5,
    "metric_id": "4e18df3d-307d-4b3f-aba4-a434ca893c8a",
    "severity": "warning",
    "metric_parameters": {
        "classifier_threshold": 0.5,
        "predicted_property": "prediction"
    },
    "filters": [
        {
            "property": "AGE",
            "comparator": "eq",
            "value": 22
        }
    ]
}


def patch_alert_rule_post(*args, **kwargs):
    alert_rule_dict["enabled"] = True  # enabled is not a param to the alert rule creation function but defaults to True
    assert kwargs['json'] == alert_rule_dict
    alert_rule_object.id = "8c298d3f-eb38-47d2-85db-182a92a84af3"
    return MockResponse(alert_rule_object, 200)


def path_alert_rule_get(*args, **kwargs):
    alert_rule_list = {
        "data": [
            {
              "name": "Lower Bound for Total Inference Count By Class - Dog",
              "metric_id": "e14e6aac-0c94-4a78-a104-871f70b8b476",
              "threshold": 200,
              "bound": "lower",
              "severity": "warning",
              "lookback_period": 360,
              "subsequent_alert_wait_time": 720,
              "enabled": True,
              "id": "2905da23-fa72-4299-92d6-40007d3a2a03",
              "metric_name": "Total Inference Count By Class - Dog",
              "attribute_name": "dog",
              "metric_parameters": {
                  "predicted_class": "dog"
              }
            },
            {
              "name": "Lower Bound for Total Inference Count By Class - Dog",
              "metric_id": "e14e6aac-0c94-4a78-a104-871f70b8b476",
              "threshold": 200,
              "bound": "lower",
              "severity": "warning",
              "lookback_period": 360,
              "subsequent_alert_wait_time": 720,
              "enabled": True,
              "id": "2905da23-fa72-4299-92d6-40007d3a2a03",
              "metric_name": "Total Inference Count By Class - Dog",
              "attribute_name": "dog"
            }
        ],
        "page": 1,
        "page_size": 0,
        "total_pages": 0,
        "total_count": 0
        }
    return MockResponse(alert_rule_list, 200)


def patch_alert_rule_patch(*args, **kwargs):
    alert_rule_dict["id"] = "123456789"
    return MockResponse(alert_rule_dict, 200)


class TestAlertRule:

    def test_alert_rule_create(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        # test successful creation of alert rule
        monkeypatch.setattr(HTTPClient, "post", patch_alert_rule_post)
        res = binary_classification_model.create_alert_rule(**alert_rule_dict)
        assert res == alert_rule_object

    def test_get_alert_rules(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        monkeypatch.setattr(HTTPClient, "get", path_alert_rule_get)
        res = binary_classification_model.get_alert_rules()
        assert len(res) == 2
        for alert_rule in res:
            assert isinstance(alert_rule, AlertRule)

    def test_alert_rule_patch(self, monkeypatch, mock_cred_env_vars, binary_classification_model: ArthurModel):
        # test updating alert rule passing id separately
        monkeypatch.setattr(HTTPClient, "patch", patch_alert_rule_patch)
        res = binary_classification_model.update_alert_rule(alert_rule_object, "123456789")
        alert_rule_object.id = "123456789"
        assert res == alert_rule_object

        # test creating alert rule using an AlertRule object
        res = binary_classification_model.update_alert_rule(alert_rule=alert_rule_object)
        alert_rule_object.id = "123456789"
        assert res == alert_rule_object

        # ensure an error is thrown if an alert rule id is not provided
        alert_rule_object.id = None
        with pytest.raises(Exception) as context:
            binary_classification_model.update_alert_rule(alert_rule_object)

        assert "alert_rule_to_update must have a valid id" in context.value.args[0]

    @pytest.mark.parametrize('metric_parameters, expected_parameters, expected_error', [
        (
            {"classifier_threshold": 0.5, "predicted_property": "prediction"},
            {"classifier_threshold": 0.5, "predicted_property": "prediction"},
            False
        ),
        (
            {"classifier_threshold": 0.5, "predicted_property": ["prediction", "prediction2"]},
            None,
            True
        ),
        (
            {"classifier_threshold": 0.5, "predicted_property": ["prediction"]},
            {"classifier_threshold": 0.5, "predicted_property": "prediction"},
            False
        ),
        (
            None,
            {},
            False
        )
    ], ids=['valid_params', 'invalid_params', 'nested_valid_params', 'null params'])
    def test_validate_parameters_for_alert(self, metric_parameters, expected_parameters, expected_error):
        if expected_error:
            with pytest.raises(ArthurUserError):
                _ = validate_parameters_for_alert(metric_parameters)
        else:
            validated_parameters = validate_parameters_for_alert(metric_parameters)
            assert validated_parameters == expected_parameters



import pytest
import responses

from arthurai.common.constants import ModelStatus, ONBOARDING_SPINNER_MESSAGE, ONBOARDING_UPDATE_MESSAGE
from arthurai.common.exceptions import ArthurInternalError
from arthurai.core.model_status_waiter import ModelStatusWaiter
from tests.helpers import mock_get
from tests.fixtures.models import regression as reg
from tests.fixtures.mocks import client

CREATING_STATUS_LIST = [ModelStatus.Ready, ModelStatus.CreationFailed]


@responses.activate
def test_model_status_wait_creating_ready(client):
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Creating
    }, status=200)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Ready
    }, status=200)

    model_status_waiter = ModelStatusWaiter(reg.TABULAR_MODEL_ID, client=client.client)

    status = model_status_waiter.wait_for_valid_model_status(CREATING_STATUS_LIST,
                                                             ONBOARDING_SPINNER_MESSAGE, ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 2
    assert status == ModelStatus.Ready


@responses.activate
def test_model_status_wait_creating_failed(client):
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.Creating
    }, status=200)
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={
        'id': reg.TABULAR_MODEL_ID,
        'model_group_id': reg.TABULAR_MODEL_GROUP_ID,
        'version_sequence_num': 1,
        'attributes': [],
        'status': ModelStatus.CreationFailed
    }, status=200)

    model_status_waiter = ModelStatusWaiter(reg.TABULAR_MODEL_ID, client=client.client)

    status = model_status_waiter.wait_for_valid_model_status(CREATING_STATUS_LIST,
                                                             ONBOARDING_SPINNER_MESSAGE, ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 2
    assert status == ModelStatus.CreationFailed


@responses.activate
def test_model_status_get_model_exception(client):
    mock_get(f"/api/v3/models/{reg.TABULAR_MODEL_ID}", response_body={}, status=500)

    model_status_waiter = ModelStatusWaiter(reg.TABULAR_MODEL_ID, client=client.client)

    with pytest.raises(ArthurInternalError):
        model_status_waiter.wait_for_valid_model_status(CREATING_STATUS_LIST,
                                                        ONBOARDING_SPINNER_MESSAGE, ONBOARDING_UPDATE_MESSAGE)
    assert len(responses.calls) == 1




import pytest

from arthurai.common.exceptions import *


def test_user_value_error_is_arthur_user_error():
    try:
        raise UserValueError("bad value")
    except ArthurUserError:
        print("expected exception caught")


def test_user_value_error_is_arthur_error():
    try:
        raise UserValueError("bad value")
    except ArthurError:
        print("expected exception caught")


def test_user_value_error_is_value_error():
    try:
        raise UserValueError("bad value")
    except ValueError:
        print("expected exception caught")


def test_internal_value_error_is_arthur_internal_error():
    try:
        raise InternalValueError("bad value")
    except ArthurInternalError:
        print("expected exception caught")


def test_internal_value_error_is_arthur_error():
    try:
        raise InternalValueError("bad value")
    except ArthurError:
        print("expected exception caught")


def test_internal_value_error_is_value_error():
    try:
        raise InternalValueError("bad value")
    except ValueError:
        print("expected exception caught")


def test_arthur_internal_error_reraised():
    @arthur_excepted("expected failure")
    def internal_error():
        raise ArthurInternalError

    with pytest.raises(ArthurInternalError):
        internal_error()


def test_arthur_user_error_reraised():
    @arthur_excepted("expected failure")
    def user_error():
        raise ArthurUserError

    with pytest.raises(ArthurUserError):
        user_error()


def test_arthur_external_error():
    @arthur_excepted("expected failure")
    def external_error():
        raise ValueError

    with pytest.raises(ArthurInternalError):
        external_error()


def test_missing_arguments():
    @arthur_excepted("expected failure")
    def my_func(pos_only, pos_or_key,  default_param=None, *, key_only):
        pass

    # missing pos_only
    with pytest.raises(MissingParameterError):
        my_func(pos_or_key=None, key_only=None)

    # missing pos_or_key
    with pytest.raises(MissingParameterError):
        my_func(None, key_only=None)

    # missing key_only
    with pytest.raises(MissingParameterError):
        my_func(None, None)

    # valid with keyword
    assert my_func(None, pos_or_key=None, key_only=None) is None

    # valid with positional
    assert my_func(None, None, key_only=None) is None



import pytest
import pandas as pd
from arthurai import ArthurAttribute, ArthurModel
from arthurai.common.constants import InputType, OutputType, ValueType, Stage
from arthurai.core.attributes import AttributeCategory
from tests.fixtures.models.multiclass import DOG_PRED, CAT_PRED, GROUND_TRUTH, UNLINKED_CAT_PRED, UNLINKED_DOG_PRED


INPUT_TEXT = ["hello there", "hi", "what's up", "yo", "cool", "cool beans"]
NLP_DATAFRAME = pd.DataFrame({"input_text": INPUT_TEXT,
                              "pred": [90, 80, 50, 31, 26, 18],
                              "gt": [95, 81, 42, 30, 30, 27]})


DOG_CLASS_STRING = "dog class"
CAT_CLASS_STRING = "cat_class"
NLP_CLASSIFICATION_DATAFRAME = pd.DataFrame({"input_text": INPUT_TEXT,
                                             DOG_PRED: [0.6, 0.8, 0.2, 0.45, 0.94, 0.12],
                                             CAT_PRED: [0.4, 0.2, 0.8, 0.65, 0.06, 0.88],
                                             GROUND_TRUTH: [DOG_CLASS_STRING, CAT_CLASS_STRING, DOG_CLASS_STRING, CAT_CLASS_STRING,
                                                            CAT_CLASS_STRING, DOG_CLASS_STRING]})

INPUT_TEXT_ATTR = ArthurAttribute(name='input_text',
                                  value_type=ValueType.Unstructured_Text,
                                  stage=Stage.ModelPipelineInput,
                                  # TODO: this seems like it should be false,
                                  #  but check that this field isn't used somewhere else where it needs to be true
                                  categorical=True,
                                  position=0)

GT_STRING_ATTR = ArthurAttribute(name=GROUND_TRUTH,
                                 value_type=ValueType.String,
                                 stage=Stage.GroundTruthClass,
                                 categorical=True,
                                 categories=[AttributeCategory(value=DOG_CLASS_STRING),
                                             AttributeCategory(value=CAT_CLASS_STRING)],
                                 position=0)

CLASSIFICATION_MODEL_ATTRIBUTES = [UNLINKED_CAT_PRED, UNLINKED_DOG_PRED, INPUT_TEXT_ATTR, GT_STRING_ATTR]

UNINITIALIZED_MODEL_DATA = {"partner_model_id": "test",
                            "client": None,
                            "input_type": InputType.NLP}


@pytest.fixture
def initial_nlp_regression_model():
    return ArthurModel(
        **UNINITIALIZED_MODEL_DATA,
        output_type=OutputType.Regression
    )


@pytest.fixture
def initial_nlp_classification_model():
    return ArthurModel(
        **UNINITIALIZED_MODEL_DATA,
        output_type=OutputType.Multiclass
    )




import logging
import sys

DEFAULT_LOG_LEVEL = logging.INFO

# define formatters

# e.g.: 17:31:27 - arthurai - info message
LITE_FORMATTER = logging.Formatter('%(asctime)s - arthurai - %(message)s', "%H:%M:%S")
# e.g.: 2022-10-13 17:31:27,703 - arthurai.core.data_service - ERROR - error message
FULL_FORMATTER = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')


# adapted from https://stackoverflow.com/a/16066513
class InfoFilter(logging.Filter):
    """
    A logging filter to only log messages that are debug or info level (i.e. never log warning or error messages)
    """
    def filter(self, rec):
        return rec.levelno in (logging.DEBUG, logging.INFO)


def initialize_logging():
    """Creates a parent 'arthurai' logger with a console output and INFO level.
    """
    # fetch logger and set default level
    arthur_logger = logging.getLogger("arthurai")
    arthur_logger.setLevel(DEFAULT_LOG_LEVEL)

    # create handlers

    # stderr handler always uses the full formatter and logs to the "standard error" stream: this shows up as red text
    #  in jupyter notebooks. it logs warning and error messages in this way
    stderr_handler = logging.StreamHandler()
    stderr_handler.setLevel(logging.WARNING)
    stderr_handler.setFormatter(FULL_FORMATTER)

    # stdout handler by default uses the lite formatter and logs to the "standard out" stream: this shows up as regular
    #  text in jupyter notebooks. it logs debug and info messages, but the debug messages don't make it through the
    #  'arthurai' logger log level set at the top of this function in practice
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(logging.DEBUG)
    stdout_handler.addFilter(InfoFilter())
    stdout_handler.setFormatter(LITE_FORMATTER)

    # add handlers to logger
    arthur_logger.handlers = []
    arthur_logger.addHandler(stdout_handler)
    arthur_logger.addHandler(stderr_handler)


def _set_arthur_log_level(level: int):
    logging.getLogger("arthurai").setLevel(level)


def enable_debug_logging():
    """Enables debug logging for the arthurai package. Note that log messages may be dropped while this function is
     being executed.
    """
    # recreate single handler to output everything to stderr
    stderr_handler = logging.StreamHandler()
    stderr_handler.setLevel(logging.DEBUG)
    stderr_handler.setFormatter(FULL_FORMATTER)

    # set new handler to be the only one
    logger = logging.getLogger("arthurai")
    logger.handlers = []
    logger.addHandler(stderr_handler)

    _set_arthur_log_level(logging.DEBUG)


def disable_debug_logging():
    """Disables debug logging for the arthurai package.  Note that log messages may be dropped while this function is
     being executed.
    """
    initialize_logging()



import logging

import pytest

from arthurai import ArthurAI


BASE_URL = "https://mock"
ACCESS_KEY = "FAKE_ACCESS_KEY"
USER_TOKEN = "fake-token"
USER_TOKEN_JAN_1_2023_MIDNIGHT_EXPIRY = \
    'eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJleHAiOjE2NzI1MzEyMDB9.BM498J3ulaG6v69S6qMEz6ETWgUeXZ7dpb-ZE4VOons'


@pytest.fixture(scope="class")
def client():
    logging.info("Creating mock ArthurAI Client")

    access_key = "FAKE_ACCESS_KEY"

    config = {
        'access_key': access_key,
        'url': BASE_URL,
        'offline': True
    }
    yield ArthurAI(config)



import pytest
from tests.fixtures.models.multiclass import DOG_GROUND_TRUTH_ATTR, INT_INPUT_ATTR
from tests.fixtures.models.regression import PRED_ATTR


@pytest.mark.parametrize('attribute, expected_str', [
    (DOG_GROUND_TRUTH_ATTR, "GROUND_TRUTH[0]:dog_gt\n\tCategorical: [0, 1]\n\tLink: dog_pred"),
    (INT_INPUT_ATTR, "PIPELINE_INPUT[0]:int_input\n\tCategorical: [1, 2]"),
    (PRED_ATTR, "PREDICTED_VALUE[0]:pred\n\tRange: [23.0, 94.2]\n\tLink: ground_truth")
], ids=['categorical with link', 'categorical no link', 'range'])
def test__str__(attribute, expected_str):
    string_rep = str(attribute)
    assert string_rep == expected_str



"""
Package for downloading data from the Arthur S3 bucket

.. warning::
    This package is experimental and is under active development. The API for this package (its specific methods and
    workflows) are subject to change as the design and scope of this package evolve.
"""
from .arthur_example import ArthurExample



import logging
import threading
import time
from http import HTTPStatus
from typing import List

from yaspin import yaspin

from arthurai.client.http.requests import HTTPClient
from arthurai.common.constants import ModelStatus
from arthurai.common.exceptions import ExpectedParameterNotFoundError, ArthurInternalError

logger = logging.getLogger(__name__)


class ModelStatusWaiter:
    def __init__(self, model_id: str, client: HTTPClient):
        self.model_id = model_id
        self._client = client
        self._status = None
        self._model_status_exception: bool = False
        self._model_status_available = threading.Event()

    def wait_for_valid_model_status(self, status_list: List, spinner_message: str,
                                    update_message: str, update_interval_seconds=60) -> ModelStatus:
        """ Wait for the model to reach a status on the status_list

        Starts the thread that is responsible for checking the status of the model until it reaches a status
        passed in the list as argument or an exception occurs, and runs a spinner providing updates while waiting for it.

        :param status_list: List of :class:`arthurai.common.constants.ModelStatus` list of valid model status
        :param spinner_message: The message you want to display when spinner is running and waiting for results
        :param update_message: The message you want to display when update_interval_seconds time is exceeded
        :param update_interval_seconds: Frequency at which tan update is provided on the console
        """

        self._model_status_available.clear()
        self._model_status_exception = False
        thread = threading.Thread(target=self._await_model_final_status_thread, args=[status_list])
        thread.start()

        with yaspin(text=spinner_message) as sp:
            while not self._model_status_available.wait(timeout=update_interval_seconds):
                sp.write(update_message)
            sp.stop()

        if self._model_status_exception is True:
            raise ArthurInternalError(
                "Failed when trying to fetch the status of the model. Please retry to wait for the status of the model.")

        return self._status


    def _await_model_final_status_thread(self, status_list: List, poll_interval_seconds=5):
        """ Thread that polls the model status

        Polls the model status until it reaches a status in the status_list or an exception occurs.
        This method is used as a thread and sets the flag model_status_available at the end of execution.
        Do not use as a regular function.
        :param status_list: List of :class:`arthurai.common.constants.ModelStatus` list of valid model status
        :param poll_interval_seconds: polling interval to check status on the model
        """
        while True:
            try:
                self._update_model_status()
            except BaseException as ex:
                logger.exception(ex)
                self._model_status_exception = True
                self._model_status_available.set()  # setting the flag on the threading event when result is available
                break
            if self._status in status_list:  # if the status is in status_list, set the flag and break
                self._model_status_available.set()  # setting the flag on the threading event when result is available
                break
            time.sleep(poll_interval_seconds)

    def _update_model_status(self):
        """Updates the latest `status` field of the current model"""
        model_resp = self._client.get(f'/models/{self.model_id}', return_raw_response=True,
                                      validation_response_code=HTTPStatus.OK)
        model_resp_json = model_resp.json()
        if 'id' not in model_resp_json:
            raise ExpectedParameterNotFoundError(
                f"An error occurred: {model_resp}, {model_resp.status_code}, {model_resp.content}")
        self._status = model_resp_json['status']


import pandas as pd
import pytest
from typing import List
from dataclasses import dataclass

from arthurai.core.attributes import ArthurAttribute
from arthurai.common.constants import ValueType, Stage
from arthurai.datasets import ArthurExample


@dataclass
class S3ExampleDataTestCase:
    example_name: str
    input_attrs: List[ArthurAttribute]
    gt_attrs: List[ArthurAttribute]
    pred_attrs: List[ArthurAttribute]
    non_input_attrs: List[ArthurAttribute]


NEW_CREDIT_CARD_DEFAULT_DRIFTY_INPUT_ATTRIBUTES = \
    [ArthurAttribute(name="LIMIT_BAL", value_type=ValueType.Integer, stage=Stage.ModelPipelineInput)] +\
    [ArthurAttribute(name=f"PAY_{i}", value_type=ValueType.Integer, stage=Stage.ModelPipelineInput) for i in
     [0, 2, 3, 4, 5, 6]] +\
    [ArthurAttribute(name=f"BILL_AMT{i}", value_type=ValueType.Integer, stage=Stage.ModelPipelineInput) for i in
     [1, 2, 3, 4, 5, 6]] +\
    [ArthurAttribute(name=f"PAY_AMT{i}", value_type=ValueType.Integer, stage=Stage.ModelPipelineInput) for i in
     [1, 2, 3, 4, 5, 6]]

NEW_CREDIT_CARD_DEFAULT_DRIFTY = S3ExampleDataTestCase(
    "new_credit_card_default_drifty",
    NEW_CREDIT_CARD_DEFAULT_DRIFTY_INPUT_ATTRIBUTES,
    [ArthurAttribute(name="ground_truth_credit_default", value_type=ValueType.Integer, stage=Stage.GroundTruthClass)],
    [ArthurAttribute(name="pred_proba_credit_default", value_type=ValueType.Float, stage=Stage.PredictedValue)],
    [ArthurAttribute(name=name, value_type=ValueType.Integer, stage=Stage.NonInputData) for name in ["ID", "MARRIAGE", "EDUCATION", "AGE", "SEX"]]
)

ELECTRICITY_INPUT_ATTRIBUTES = [
    ArthurAttribute(name=name, value_type=ValueType.Float, stage=Stage.ModelPipelineInput)
    for name in ["date", "period", "nswprice", "nswdemand", "vicprice", "vicdemand", "transfer"]] + [
    ArthurAttribute(name="day", value_type=ValueType.Integer, stage=Stage.ModelPipelineInput)
]

ELECTRICITY = S3ExampleDataTestCase(
    "electricity",
    ELECTRICITY_INPUT_ATTRIBUTES,
    [ArthurAttribute(name="price_increase", value_type=ValueType.Integer, stage=Stage.GroundTruthClass)],
    [ArthurAttribute(name="pred_price_increase", value_type=ValueType.Float, stage=Stage.PredictedValue)],
    []
)

BOSTON_HOUSING_INPUT_ATTRIBUTES = [
    ArthurAttribute(name=name, value_type=ValueType.Float, stage=Stage.ModelPipelineInput)
    for name in ["crim", "zn", "indus", "nox", "rm", "age", "dis", "ptratio", "bb", "lstat"]] + [
    ArthurAttribute(name=name, value_type=ValueType.Integer, stage=Stage.ModelPipelineInput)
    for name in ["chas", "rad", "tax"]
]

BOSTON_HOUSING = S3ExampleDataTestCase(
    "boston_housing",
    BOSTON_HOUSING_INPUT_ATTRIBUTES,
    [ArthurAttribute(name="medv", value_type=ValueType.Float, stage=Stage.GroundTruth)],
    [ArthurAttribute(name="pred_medv", value_type=ValueType.Float, stage=Stage.PredictedValue)],
    []
)

EXAMPLE_DATASETS = [
    NEW_CREDIT_CARD_DEFAULT_DRIFTY,
    ELECTRICITY,
    BOSTON_HOUSING
]


@pytest.mark.parametrize("example_dataset", EXAMPLE_DATASETS)
def test_s3_example(example_dataset: S3ExampleDataTestCase):
    example_name = example_dataset.example_name
    input_attributes = example_dataset.input_attrs
    gt_attributes = example_dataset.gt_attrs
    pred_attributes = example_dataset.pred_attrs
    non_input_attributes = example_dataset.non_input_attrs
    arthur_example = ArthurExample(example_name)
    attributes = input_attributes + gt_attributes + pred_attributes + non_input_attributes

    # assert feature name properties are correct
    assert set(arthur_example.input_attribute_names) == set([attr.name for attr in input_attributes])
    assert set(arthur_example.gt_attribute_names) == set([attr.name for attr in gt_attributes])
    assert set(arthur_example.pred_attribute_names) == set([attr.name for attr in pred_attributes])
    assert set(arthur_example.non_input_attribute_names) == set([attr.name for attr in non_input_attributes])

    # validate that the ArthurExample methods split data correctly by stage, by train-test split, and by datatype
    validate_arthur_example_method(arthur_example, arthur_example.get_dataset, attributes)
    validate_arthur_example_method(arthur_example, arthur_example.get_inputs, input_attributes)
    validate_arthur_example_method(arthur_example, arthur_example.get_ground_truth_data, gt_attributes)
    validate_arthur_example_method(arthur_example, arthur_example.get_predictions, pred_attributes)
    validate_arthur_example_method(arthur_example, arthur_example.get_non_input_data, non_input_attributes)


# run tests on a method of an ArthurExample object
# each method returns a dataframe which should correspond to the expected_attrs
def validate_arthur_example_method(arthur_example, arthur_example_method, expected_attrs: List[ArthurAttribute]):
    if not expected_attrs:
        with pytest.raises(ValueError):
            arthur_example_method()
    else:
        df = arthur_example_method()
        ref_df, inf_df = arthur_example_method(split=True)
        ref_df_even, inf_df_even = arthur_example_method(split=True, test_split_size=0.5)
        assert set(df.columns) == set([attr.name for attr in expected_attrs])
        assert not df.isnull().values.any()
        type_check_arthur_example_method_result(df, expected_attrs)
        split_check_arthur_example_method_result(arthur_example, df, ref_df, inf_df, ref_df_even, inf_df_even)


# assert that the result of the arthur_example method type-checks according to the attributes
def type_check_arthur_example_method_result(result_data: pd.DataFrame, attributes: List[ArthurAttribute]):
    for c in result_data.columns:
        attrs = [attr for attr in attributes if attr.name == c]
        assert len(attrs) == 1
        attr = attrs[0]
        if attr.value_type == ValueType.Integer:
            assert result_data[c].dtype == int
        elif attr.value_type == ValueType.Float:
            assert result_data[c].dtype == float
        elif attr.value_type == ValueType.Boolean:
            assert result_data[c].dtype == bool
        elif attr.value_type in [ValueType.String, ValueType.Unstructured_Text, ValueType.Image]:
            assert result_data[c].dtype == str


# assert that the result of the arthur_example method is train/test split correctly
def split_check_arthur_example_method_result(
        arthur_example: ArthurExample,
        result_data: pd.DataFrame,
        ref: pd.DataFrame,
        inf: pd.DataFrame,
        ref_even: pd.DataFrame,
        inf_even: pd.DataFrame
):
    default_num_ref = int((1 - arthur_example.example_schema.default_data_split_test_size) * len(result_data))
    default_num_inf = int(arthur_example.example_schema.default_data_split_test_size * len(result_data))
    assert default_num_ref - 1 <= len(ref) <= default_num_ref + 1
    assert default_num_inf - 1 <= len(inf) <= default_num_inf + 1
    assert len(result_data) // 2 - 1 <= len(ref_even) <= len(result_data) // 2 + 1
    assert len(result_data) // 2 - 1 <= len(inf_even) <= len(result_data) // 2 + 1


def test_bad_example_name():
    with pytest.raises(ValueError):
        ArthurExample("asdfjasdgkfj")



# arthur_example.py
from dataclasses import dataclass
import pandas as pd
from pathlib import Path
from sklearn.model_selection import train_test_split
from typing import List, Union, Optional, Tuple
from arthurai.core.base import ArthurBaseJsonDataclass
from arthurai.core.models import ArthurModel
from arthurai.common.constants import Stage
from arthurai.common.exceptions import ArthurDownloadError
from arthurai.datasets.download import ArthurExampleDownloader, default_download_directory, load_downloaded_file


@dataclass
class ArthurExampleSchema(ArthurBaseJsonDataclass):
	"""
	Class for the schema of all the data used in an Arthur Example
	"""

	arthur_model: ArthurModel
	default_data_split_test_size: Optional[float] = 0.3
	default_random_state: int = 278487


class ArthurExample:
	"""
	Class for a user to interface with example data for model analysis and using Arthur
	"""

	def __init__(self, name: str, download_destination_folder: Optional[Path] = None):
		"""
		Constructs an ArthurExample

		:param name: str, the name of the example
		:param download_destination_folder: Path, the destination to save files to. Defaults to current directory if
			none is specified.
		"""
		if download_destination_folder is None:
			download_destination_folder = Path.cwd() / default_download_directory
		try:
			downloader = ArthurExampleDownloader(name, download_destination_folder)
		except ArthurDownloadError:
			raise ValueError(f"Cannot download data: {name} is not the name of a valid ArthurExample.") from None

		self.example_schema = ArthurExampleSchema.from_dict(downloader.download_metadata())
		self.dataset = load_downloaded_file(downloader.download_dataset())
		self.predictions = load_downloaded_file(downloader.download_pretrained_model_predictions())

	def get_dataset(self, split: bool = False, test_split_size: Optional[float] = None) \
		-> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
		"""
		Returns a dataframe of the model's inputs, predictions, ground truth labels, and non input data

		:param split: bool, whether to return the data split into train/test.
		:param test_split_size: the percentage of data to be split into the test set if splitting the data. If None,
			uses the default test split defined in the example schema.
		:return: dataframe of inputs, predictions, ground truth labels, and non input data
		"""
		if split:
			train_inputs, test_inputs = self.get_inputs(split=True, test_split_size=test_split_size)
			train_pred, test_pred = self.get_predictions(split=True, test_split_size=test_split_size)
			train_gt, test_gt = self.get_ground_truth_data(split=True, test_split_size=test_split_size)
			train_dataset = pd.concat([train_inputs, train_pred, train_gt], axis=1)
			test_dataset = pd.concat([test_inputs, test_pred, test_gt], axis=1)
			if len(self.non_input_attribute_names) > 0:
				train_non_input, test_non_input = self.get_non_input_data(split=True, test_split_size=test_split_size)
				train_dataset = pd.concat([train_dataset, train_non_input], axis=1)
				test_dataset = pd.concat([test_dataset, test_non_input], axis=1)
			return train_dataset, test_dataset
		else:
			dataset = pd.concat([self.get_inputs(), self.get_predictions(), self.get_ground_truth_data()], axis=1)
			if len(self.non_input_attribute_names) > 0:
				dataset = pd.concat([dataset, self.get_non_input_data()], axis=1)
			return dataset

	def get_inputs(self, split: bool = False, test_split_size: Optional[float] = None, random_state: Optional[int] = None) \
		-> Union[pd.DataFrame, pd.Series, Tuple[pd.DataFrame, pd.DataFrame], Tuple[pd.Series, pd.Series]]:
		"""
		Returns the model's input attribute feature values from self.dataset

		:param split: bool, whether to return the data split by reference/inference (AKA train/test),
			or whether to return a single dataframe
		:param test_split_size: the percentage of data to be split into the test set if splitting the data
		:param random_state: int, random state for optional split
		:return: dataframe of model input features
		"""
		X = self.dataset[self.input_attribute_names]
		if split:
			if test_split_size is None: test_split_size = self.example_schema.default_data_split_test_size
			if random_state is None: random_state = self.example_schema.default_random_state
			return train_test_split(X, test_size=test_split_size, random_state=random_state)
		return X

	def get_ground_truth_data(self, split: bool = False, test_split_size: Optional[float] = None, random_state: Optional[int] = None) \
		-> Union[pd.DataFrame, pd.Series, Tuple[pd.DataFrame, pd.DataFrame], Tuple[pd.Series, pd.Series]]:
		"""
		Returns the model's ground truth labels from self.dataset

		:param split: bool, whether to return the data split by reference/inference (AKA train/test),
			or whether to return a single dataframe
		:param test_split_size: the percentage of data to be split into the test set if splitting the data
		:param random_state: int, random state for optional split
		:return: dataframe of model ground truth values
		"""
		y = self.dataset[self.gt_attribute_names]
		if split:
			if test_split_size is None: test_split_size = self.example_schema.default_data_split_test_size
			if random_state is None: random_state = self.example_schema.default_random_state
			return train_test_split(y, test_size=test_split_size, random_state=random_state)
		return y

	def get_predictions(self, split: bool = False, test_split_size: Optional[float] = None, random_state: Optional[int] = None) \
		-> Union[pd.DataFrame, pd.Series, Tuple[pd.DataFrame, pd.DataFrame], Tuple[pd.Series, pd.Series]]:
		"""
		Returns the model's predicted values from self.predictions

		:param split: bool, whether to return the data split by reference/inference (AKA train/test),
			or whether to return a single dataframe
		:param test_split_size: the percentage of data to be split into the test set if splitting the data
		:param random_state: int, random state for optional split
		:return: dataframe of model prediction data
		"""
		y_pred = self.predictions[self.pred_attribute_names]
		if split:
			if test_split_size is None: test_split_size = self.example_schema.default_data_split_test_size
			if random_state is None: random_state = self.example_schema.default_random_state
			return train_test_split(y_pred, test_size=test_split_size, random_state=random_state)
		return y_pred

	def get_non_input_data(self, split: bool = False, test_split_size: Optional[float] = None, random_state: Optional[int] = None) \
		-> Union[pd.DataFrame, pd.Series, Tuple[pd.DataFrame, pd.DataFrame], Tuple[pd.Series, pd.Series]]:
		"""
		Returns the model's non input attributes from self.dataset

		:param split: bool, whether to return the data split by reference/inference (AKA train/test),
			or whether to return a single dataframe
		:param test_split_size: the percentage of data to be split into the test set if splitting the data
		:param random_state: int, random state for optional split
		:return: dataframe of model non-input attribute data
		"""
		if len(self.non_input_attribute_names) > 0:
			non_input_data = self.dataset[self.non_input_attribute_names]
			if split:
				if test_split_size is None: test_split_size = self.example_schema.default_data_split_test_size
				if random_state is None: random_state = self.example_schema.default_random_state
				return train_test_split(non_input_data, test_size=test_split_size, random_state=random_state)
			return non_input_data
		raise ValueError("This example has no non_input attributes.")

	@property
	def input_attribute_names(self) -> List[str]:
		"""
		The list of feature names which make up the model's input attributes

		:return: list of input attribute names
		"""
		return [attr.name for attr in self.example_schema.arthur_model.get_attributes(stage=Stage.ModelPipelineInput)]

	@property
	def gt_attribute_names(self) -> List[str]:
		"""
		The list of attribute names which make up the model's ground truth attributes
		Either GroundTruth or GroundTruthClass

		:return: list of ground truth attribute names
		"""
		gt_attrs = [attr.name for attr in self.example_schema.arthur_model.get_attributes(stage=Stage.GroundTruth)]
		if len(gt_attrs) > 0: return gt_attrs
		return [attr.name for attr in self.example_schema.arthur_model.get_attributes(stage=Stage.GroundTruthClass)]

	@property
	def pred_attribute_names(self) -> List[str]:
		"""
		The list of feature names which make up the model's predicted attributes

		:return: list of predicted attribute names
		"""
		return [attr.name for attr in self.example_schema.arthur_model.get_attributes(stage=Stage.PredictedValue)]

	@property
	def non_input_attribute_names(self) -> List[str]:
		"""
		The list of feature names which make up the model's non-input attributes

		:return: list of non-input attribute names
		"""
		return [attr.name for attr in self.example_schema.arthur_model.get_attributes(stage=Stage.NonInputData)]



from http import HTTPStatus
from typing import Optional, Tuple, Any, Dict
import requests
from urllib.parse import urlparse
import logging

from arthurai.client.validation import validate_response_status
from arthurai.common.exceptions import UserValueError, ArthurInternalError, ForbiddenError
from arthurai.core.auth_info import AuthInfo

JSON_CONTENT_TYPE = "application/json"

logger = logging.getLogger(__name__)


def construct_url(*parts: str, validate=True, default_https=True) -> str:
    """Construct a url from various parts

    Useful for joining pieces which may or may not have leading and/or trailing
    slashes. e.g. construct_url("http://arthur.ai/", "/api/v3", "/users") will yield the same valid url as
    construct_url("http://arthur.ai", "api/v3/", "users/").

    :param validate: if True, validate that the URL is valid
    :param default_https: if True, allow urls without a scheme and use https by default
    :param parts: strings from which to construct the url
    :return: a fully joined url
    """
    # join parts
    url = '/'.join(s.strip('/') for s in parts)

    # add scheme
    parsed_url = urlparse(url)
    if parsed_url.scheme is None or parsed_url.scheme == "":
        if default_https:
            logger.warning("No url scheme provided, defaulting to https")
            url = "https://" + url
            parsed_url = urlparse(url)
        elif validate:
            raise UserValueError(f"No scheme provided in URL {url}")

    # validate
    if validate and (parsed_url.scheme is None or parsed_url.scheme == ""
                     or parsed_url.netloc is None or parsed_url.netloc == ""):
        joiner = "', '"
        raise UserValueError(f"Invalid url, cannot construct URL from parts '{joiner.join(parts)}'")

    return url


def get_current_org(api_http_host: str, api_prefix: str, access_key: str, verify_ssl: bool = True) \
        -> Optional[str]:
    """Get the current organization for the provided access key

    :param api_http_host: base url of the host to connect to, including protocol (e.g. "https://app.arthur.ai")
    :param api_prefix: prefix of the API to connect to (e.g. "/api/v3")
    :param access_key: API Key to pass to the API
    :param verify_ssl: Boolean for whether requests should verify that the SSL certificate is valid and not self-signed
    :return: the organization ID associated with the provided access key, None if no such organization exists
    """
    try:
        internal_user_org = get_arthur_internal_user_org(api_http_host, api_prefix, access_key, verify_ssl)
    except ForbiddenError:
        auth_info = get_auth_info(api_http_host, api_prefix, access_key, verify_ssl)
        if auth_info is None:
            raise ArthurInternalError("Invalid / non-existent auth_info")
        elif len(auth_info.organization_ids) == 1:
            # If the user is authenticated into a single org, automatically select it as the user's current org
            return auth_info.organization_ids[0]
        elif len(auth_info.organization_ids) > 1:
            # Raise exception and give the user options of available current organizations
            authenticated_org_id_str = "".join(f"{org_id}\n" for org_id in auth_info.organization_ids)[:-1]
            raise UserValueError(f"Your access_key provides access to multiple organizations - please specify one of the following:\n{authenticated_org_id_str}")
        else: # len(auth_info.organization_ids) == 0
            return None

    # Valid internal_user_org was found
    return internal_user_org


def get_auth_info(api_http_host: str, api_prefix: str, access_key: str, verify_ssl: bool = True) \
        -> Optional[AuthInfo]:
    """Get the AuthInfo struct associated with the provided access key

    :param api_http_host: base url of the host to connect to, including protocol (e.g. "https://app.arthur.ai")
    :param api_prefix: prefix of the API to connect to (e.g. "/api/v3")
    :param access_key: API token to pass to the API
    :param verify_ssl: Boolean for whether requests should verify that the SSL certificate is valid and not self-signed
    :return: the AuthInfo associated with the provided access key
    :permissions: N/A
    """
    url = construct_url(api_http_host, api_prefix, "/users/me/auth_info")
    headers = {'Authorization': access_key,
               'Accept': JSON_CONTENT_TYPE}

    try:
        resp = requests.get(url, headers=headers, verify=verify_ssl)
    except requests.exceptions.SSLError as e:
        raise UserValueError(f"""SSL Error connecting to {api_http_host}, please connect to a secure server or use 
                             verify_ssl=False to override security checks""") from e
    except requests.RequestException as e:
        raise UserValueError(f"Failed to connect to {api_http_host}, please ensure the URL is correct") from e
    if resp.status_code == HTTPStatus.UNAUTHORIZED:
        raise UserValueError("Unauthorized, please ensure your access key is correct")
    if resp.status_code == HTTPStatus.NOT_FOUND:
        raise ArthurInternalError("Auth Info endpoint not implemented by api-host")

    validate_response_status(resp, HTTPStatus.OK)
    return AuthInfo.from_json(resp.text)


def get_arthur_internal_user_org(api_http_host: str, api_prefix: str, access_key: str, verify_ssl: bool = True) \
        -> Optional[str]:
    """Get the current organization for the provided Arthur access key belonging to an Arthur internal user

    :param api_http_host: base url of the host to connect to, including protocol (e.g. "https://app.arthur.ai")
    :param api_prefix: prefix of the API to connect to (e.g. "/api/v3")
    :param access_key: API Key to pass to the API
    :param verify_ssl: Boolean for whether requests should verify that the SSL certificate is valid and not self-signed
    :return: the organization ID associated with the provided access key, None if no such organization exists
    :permissions: N/A
    """
    url = construct_url(api_http_host, api_prefix, "/users/me")
    headers = {'Authorization': access_key,
               'Accept': JSON_CONTENT_TYPE}

    try:
        resp = requests.get(url, headers=headers, verify=verify_ssl)
    except requests.exceptions.SSLError as e:
        raise UserValueError(f"""SSL Error connecting to {api_http_host}, please connect to a secure server or use 
                             verify_ssl=False to override security checks""") from e
    except requests.RequestException as e:
        raise UserValueError(f"Failed to connect to {api_http_host}, please ensure the URL is correct") from e
    if resp.status_code == HTTPStatus.UNAUTHORIZED:
        raise UserValueError("Unauthorized, please ensure your access key is correct")
    if resp.status_code == HTTPStatus.FORBIDDEN:
        # The only case in which /users/me would return a Forbidden error response is if the calling user
        # is not an Arthur internal user (e.g. using a 3rd party identity provider)
        raise ForbiddenError("Caller is not an Arthur internal user")
    validate_response_status(resp, HTTPStatus.OK)
    response_body = resp.json()
    return response_body.get("organization_id", None)


def user_login(api_http_host: str, api_prefix: str, login: str, password: str, verify_ssl: bool = True) \
        -> Tuple[str, Dict[str, Any]]:
    """Get the current organization for the provided access key

    :param api_http_host: base url of the host to connect to, including protocol (e.g. "https://app.arthur.ai")
    :param api_prefix: prefix of the API to connect to (e.g. "/api/v3")
    :param login: the username or password to use to log in
    :param password: password for the user
    :param verify_ssl: Boolean for whether requests should verify that the SSL certificate is valid and not self-signed
    :return: a tuple of (access_key, {user object})
    """
    url = construct_url(api_http_host, api_prefix, "/login")
    headers = {'Accept': JSON_CONTENT_TYPE}
    body = {'login': login,
            'password': password}
    try:
        resp = requests.post(url, headers=headers, json=body, verify=verify_ssl)
    except requests.exceptions.SSLError as e:
        raise UserValueError(f"""SSL Error connecting to {api_http_host}, please connect to a secure server or use 
                             verify_ssl=False to override security checks""") from e
    except requests.RequestException as e:
        raise UserValueError(f"Failed to connect to {api_http_host}, please ensure the URL is correct") from e
    if resp.status_code == HTTPStatus.UNAUTHORIZED:
        raise UserValueError(f"Unauthorized, please ensure your username and password are correct")

    validate_response_status(resp, HTTPStatus.OK)

    return resp.cookies.get('Authorization'), resp.json()



from datetime import timedelta
from typing import Optional, List, Union, Dict, Callable, Tuple
import os
from distutils.util import strtobool
from platform import system
from http import HTTPStatus
from datetime import datetime
import logging
from getpass import getpass

from arthurai.client.auth import AuthRefresher
from arthurai.client.helper import get_current_org
from arthurai.client.http.arthur import FALSE_OFFLINE_DEFAULT
from arthurai.client.http.requests import HTTPClient
from arthurai.common.constants import InputType, OutputType, TextDelimiter, Role, DEFAULT_SERVICE_ACCOUNT, API_PREFIX
from arthurai.common.exceptions import MissingParameterError, UserTypeError, UserValueError, arthur_excepted
from arthurai.core.models import ArthurModel, ArthurModelGroup
from arthurai.client.validation import validate_response_status
from arthurai.version import __version__ as arthur_version
from arthurai.client.helper import user_login

UNKNOWN_ORG_ID = "unknown"
ORG_ID_BODY = "organization_id"
ORG_ID_HEADER = "Arthur-Organization-ID"

logger = logging.getLogger(__name__)


def new_requests_client(access_key: Optional[str] = None,
                        url: Optional[str] = None,
                        verify_ssl: bool = True,
                        offline: bool = False,
                        organization_id: Optional[str] = None,
                        allow_insecure: bool = False,
                        header_refresh_func: Optional[Callable[[],
                                                      Tuple[Dict[str, str], timedelta]]] = None) -> HTTPClient:
    if access_key is None:
        access_key = os.getenv('ARTHUR_API_KEY')
    if url is None:
        url = os.getenv('ARTHUR_ENDPOINT_URL')

    if url is None or access_key is None:
        raise MissingParameterError("Please set api key and url either via environment variables "
                                    "(`ARTHUR_API_KEY` and `ARTHUR_ENDPOINT_URL`) or by passing parameters "
                                    "`access_key` and `url`.")

    if offline and organization_id is not None:
        raise UserValueError("You cannot specify an organization ID if you are offline.")

    if organization_id is None:
        organization_id = os.getenv('ARTHUR_ORGANIZATION_ID', None)

    if offline:
        org_id = UNKNOWN_ORG_ID
    elif organization_id is not None:
        org_id = organization_id
    else:
        org_id = get_current_org(url, API_PREFIX, access_key, verify_ssl=verify_ssl)
    user_agent = f"arthur-sdk/{arthur_version} (system={system()})"
    headers = {
        'Accept': "application/json",
        'Authorization': access_key,
        'User-Agent': user_agent,
        ORG_ID_HEADER: org_id
    }
    return HTTPClient(base_url=url, path_prefix=API_PREFIX, default_headers=headers, verify_ssl=verify_ssl,
                      allow_insecure=allow_insecure, header_refresh_func=header_refresh_func)


class ArthurAI(object):
    """A client that interacts with Arthur's servers.
    """

    def __init__(self, config=None, verify_ssl=None, url=None, access_key=None, offline=FALSE_OFFLINE_DEFAULT,
                 login: str = None, password: str=None, organization_id: Optional[str] = None,
                 allow_insecure: bool = False):

        verify_ssl_env_var = os.getenv('VERIFY_SSL')
        if verify_ssl is not None:
            if not isinstance(verify_ssl, bool):
                raise UserTypeError("verify_ssl must be of type bool")
            self.verify_ssl = verify_ssl
        elif verify_ssl_env_var:
            try:
                self.verify_ssl = bool(strtobool(verify_ssl_env_var))
            except ValueError:
                raise UserValueError("Environment Variable VERIFY_SSL must be one of True, true, T, t, False, false, F, f, 1, 0")
        else:
            self.verify_ssl = True

        # "config" dict with params for backwards compatibility
        if config:
            self.client = new_requests_client(verify_ssl=self.verify_ssl, **config)
            return

        # populate login/password from environment if supplied
        login_env = os.getenv('ARTHUR_LOGIN')
        if login is None and login_env is not None:
            login = login_env
        password_env = os.getenv('ARTHUR_PASSWORD')
        if password is None and password_env is not None:
            password = password_env

        elif login is not None:
            # if password not supplied, get it from input
            if password is None:
                password = getpass(f"Please enter password for {login}: ")

            # Get session token from login and password
            auth_token, _ = user_login(api_http_host=url, api_prefix=API_PREFIX, login=login, password=password,
                                       verify_ssl=self.verify_ssl)
            # create an auth refresher
            auth_refresher = AuthRefresher(url=url, login=login, password=password, verify_ssl=verify_ssl)
            # create the client
            self.client = new_requests_client(access_key=auth_token, url=url, verify_ssl=self.verify_ssl,
                                              offline=offline, organization_id=organization_id,
                                              allow_insecure=allow_insecure, header_refresh_func=auth_refresher.refresh)
        else:
            self.client = new_requests_client(access_key=access_key, url=url, verify_ssl=self.verify_ssl,
                                              offline=offline, organization_id=organization_id,
                                              allow_insecure=allow_insecure)

    def model(self,
              partner_model_id: Optional[str] = None,
              input_type: Optional[InputType] = None,
              output_type: Optional[OutputType] = None,
              model_type: Optional[OutputType] = None,
              display_name: Optional[str] = None,
              description: Optional[str] = None,
              tags: Optional[List[str]] = None,
              classifier_threshold: Optional[float] = None,
              is_batch: bool = False,
              text_delimiter: Optional[TextDelimiter] = None,
              expected_throughput_gb_per_day: Optional[int] = None,
              pixel_height: Optional[int] = None,
              pixel_width: Optional[int] = None) -> ArthurModel:
        """Create a new multistage model.

        :param partner_model_id: The string external id of the model. If display_name is provided without partner_model_id,
            we will generate a unique partner_model_id.
        :param input_type: the :py:class:`.InputType`
        :param output_type: the :py:class:`.OutputType`
        :param model_type: .. deprecated:: version 2.0.0 Use `output_type` instead.
        :param display_name: Optional name to display on dashboard, will default to the external id
        :param description: Optional description for the model
        :param tags: A list of string tags to associate with the model
        :param classifier_threshold: For binary classification models this is the threshold to determine a positive class, defaults to 0.5
        :param is_batch: boolean value which signifies whether the model sends inferences by batch or streaming
        :param text_delimiter: TextDelimiter used in NLP models to split documents into tokens for explanations
        :param expected_throughput_gb_per_day: Expected amount of throughput. Used to provision resources
        :param pixel_height: Image height in pixels. Needed for CV models which require images to be one size
        :param pixel_width: Image width in pixels. Needed for CV models which require images to be one size

        :return: An :py:class:`~arthurai.client.apiv3.model.ArthurModel`
        """
        if input_type is None:
            # marked as an optional parameter in the signature to maintain order of parameters
            raise MissingParameterError("input_type must be specified.")

        if output_type is None and model_type is None:
            raise MissingParameterError("Either 'output_type' or 'model_type' parameter must be specified")
        output_type = output_type if output_type is not None else model_type

        if partner_model_id is None and display_name is None:
            raise MissingParameterError("Either 'partner_model_id' or 'display_name' parameter must be specified" )
        elif partner_model_id is None and display_name is not None:
            partner_model_id = display_name.replace(" ", "_") + "_" + datetime.now().strftime('%Y-%m-%d--%H-%M-%S')
        elif display_name is None:
            display_name = partner_model_id

        return ArthurModel(
            client=self.client,
            partner_model_id=partner_model_id,
            input_type=input_type,
            output_type=output_type,
            display_name=display_name,
            description=description,
            tags=tags,
            classifier_threshold=classifier_threshold,
            is_batch=is_batch,
            text_delimiter=text_delimiter,
            expected_throughput_gb_per_day=expected_throughput_gb_per_day,
            pixel_height=pixel_height,
            pixel_width=pixel_width,
        )

    @arthur_excepted("failed to retrieve model")
    def get_model(self, identifier: str, id_type: str = 'id') -> ArthurModel:
        """Retrieve an existing model by id

        :param: identifier: Id to get the model by
        :param: id_type: Type of id the identifier is, possible options are ['id', 'partner_model_id']
        :raises ArthurUserError: failed due to user error
        :raises ArthurInternalError: failed due to an internal error
        :permissions: model read
        """
        if id_type in ['partner_model_id', 'id']:
            endpoint = f'/models/{identifier}?id_type={id_type}&expand=attributes'
        else:
            raise UserValueError(f"Invalid id_type: {id_type}, must be one of the following ['id', 'partner_model_id]")

        resp = self.client.get(endpoint, return_raw_response=True, validation_response_code=HTTPStatus.OK)

        # accuracy data is only used by the UI not needed in the SDK
        model_data = resp.json()
        if "accuracy_enabled" in model_data:
            del model_data["accuracy_enabled"]

        model = ArthurModel.from_dict(model_data)
        model._update_client(self.client)
        return model

    @arthur_excepted("failed to retrieve model group")
    def get_model_group(self, id_or_model: Union[str, ArthurModel]) -> ArthurModelGroup:
        """(BETA) Retrieve an existing model group by id

        :param: identifier: Id to get the model group by
        :raises ArthurUserError: failed due to user error
        :raises ArthurInternalError: failed due to an internal error
        :permissions: model_group read
        """
        if isinstance(id_or_model, str):
            return ArthurModelGroup._get_model_group(id_or_model, self.client)
        elif isinstance(id_or_model, ArthurModel):
            if id_or_model.model_group_id is not None:
                return ArthurModelGroup._get_model_group(id_or_model.model_group_id, self.client)
            else:
                raise UserValueError("ArthurModel provided has no assigned model_group_id.")
        else:
            raise UserTypeError("Provided neither a string ID or ArthurModel instance.")

    @arthur_excepted("failed to get service accounts")
    def get_service_accounts(self, description: Optional[str] = None):
        """Retrieve service accounts

        :param description: an optional description that returned account should match
        :return: parsed response containing a list of service accounts, in the following format:

            .. code-block:: python

                [
                    {
                        "id": "<service account id>",
                        "organization_id": "<organization id>",
                        "description": "<service account description>",
                        "token": "<service account api key>"
                    }
                ]

        """
        endpoint = "/service_accounts"
        params = None if description is None else {'description': description}
        resp = self.client.get(endpoint=endpoint, params=params, return_raw_response=True)
        validate_response_status(resp, expected_status_code=HTTPStatus.OK)

        return resp.json()['data']

    @arthur_excepted("failed to create service account")
    def create_service_account(self, description: str, role: Role):
        """Create a new service account

        :param description: the description for the new service account
        :param role: the Role of the new service account
        :return: parsed response in the following format:

            .. code-block:: python

                {
                    "id": "<service account id>",
                    "organization_id": "<organization id>",
                    "description": "<service account description>",
                    "token": "<service account api key>"
                }
        """
        endpoint = "/service_accounts"
        body = {
            "description": description,
            "role": role
        }
        resp = self.client.post(endpoint=endpoint, json=body, return_raw_response=True)
        validate_response_status(resp, expected_status_code=HTTPStatus.CREATED)

        return resp.json()

    @arthur_excepted("failed to retrieve current organization")
    def get_current_org(self) -> Dict[str,str]:
        """Retrieves the current organization that the client is operating in

        :return: a string with the UUID of the current organization
        :permissions: N/A
        """
        endpoint = "/organizations/current"
        response = self.client.get(endpoint, return_raw_response=True)
        validate_response_status(response, expected_status_code=HTTPStatus.OK)
        return response.json()

    @arthur_excepted("failed to set the current organization")
    def set_current_org(self, org_id: str) -> None:
        """Sets the current organization that the client's requests apply to

        :param org_id: The ID of the organization to set
        :return: None
        :permissions: N/A
        """
        # Set the current_org with a PUT request which returns a cookie which gets saved on the HTTPClient session
        endpoint = "/organizations/current"
        body = {
            "organization_id": org_id
        }
        response = self.client.put(endpoint, json=body, return_raw_response=True)
        validate_response_status(response, expected_status_code=HTTPStatus.OK)

        # Update the client headers to use the correct org_id for consistency and debugging
        user_agent = f"arthur-sdk/{arthur_version} (system={system()})"
        updated_headers = {
            "User-Agent": user_agent,
            ORG_ID_HEADER: org_id
        }
        self.client.session.headers.update(updated_headers)



import os
import re
from datetime import timedelta
from http import HTTPStatus
from unittest import mock

import pytest
from arthurai.core.models import ArthurModel
import responses

from tests import MockResponse

from arthurai import ArthurAI
from arthurai.common.constants import InputType, OutputType
from arthurai.client.validation import validate_multistatus_response_and_get_failures, validate_response_status
from arthurai.common.exceptions import ResponseServerError, ResponseClientError, ResponseRedirectError, \
    InternalValueError, UserValueError
from tests.fixtures.mocks import ACCESS_KEY, BASE_URL, USER_TOKEN
from tests.helpers import mock_get, mock_post


class MockAuthRefresher:

    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def refresh():
        # tell to sleep for 60 sec before refreshing
        return {'Authorization': USER_TOKEN}, timedelta(seconds=60)


class TestClient:

    def _mock_get_user(self):
        mock_get(f"/api/v3/users/me", status=HTTPStatus.OK,
                 response_body={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})

    def _mock_get_user_unauthorized(self):
        mock_get(f"/api/v3/users/me", status=HTTPStatus.UNAUTHORIZED,
                 response_body={"bad": "user"})

    def _mock_user_login(self):
        mock_post(f"/api/v3/login", status=HTTPStatus.OK, response_body={},
                  headers={"Set-Cookie": 'Authorization=' + USER_TOKEN})

    @pytest.mark.usefixtures('mock_cred_env_vars')
    def test_client_init_env_vars(self):
        client = ArthurAI(offline=True, allow_insecure=True)
        assert client.client.api_base_url == 'https://mock/api/v3'
        assert client.client.session.headers.get('Authorization') == 'access_key'
        assert client.client.session.verify is False

        client = ArthurAI(url="this is a url", access_key="this is an access key", offline=True, allow_insecure=True)
        assert client.client.api_base_url == "https://this is a url/api/v3"
        assert client.client.session.headers.get('Authorization') == "this is an access key"
        assert client.client.session.verify is False

        client = ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True, allow_insecure=True)
        assert client.client.api_base_url == "http://this is a url/api/v3"
        assert client.client.session.headers.get('Authorization') == "this is an access key"
        assert client.client.session.verify is False

        with pytest.raises(ValueError):
            ArthurAI(url="http://this is a url/apath", access_key="this is an access key", offline=True, allow_insecure=True)

        with pytest.raises(ValueError):
            ArthurAI(url="http://this is a url?food=good", access_key="this is an access key", offline=True, allow_insecure=True)

        with pytest.raises(ValueError):
            ArthurAI(url="fatfinger://this is a url", access_key="this is an access key", offline=True, allow_insecure=True)

        with pytest.raises(ValueError):
            ArthurAI(url="https:", access_key="this is an access key", offline=True, allow_insecure=True)

        with pytest.raises(ValueError):
            ArthurAI(url='http://asdfa', access_key='this is an access_key', allow_insecure=False)

    @pytest.mark.usefixtures('mock_cred_env_vars_ssl_true')
    def test_client_init_env_vars_ssl_true(self):
        client = ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True, allow_insecure=True)
        assert client.client.session.verify is True

    @pytest.mark.usefixtures('mock_cred_env_vars_ssl_garbage')
    def test_client_init_env_vars_ssl_garbage(self):
        with pytest.raises(UserValueError):
            ArthurAI(url="http://this is a url", access_key="this is an access key", offline=True)

    @responses.activate
    def test_client_init(self):
        self._mock_get_user()

        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1
        expectedResponse = {'organization_id': '65624ee6-9a73-4fe3-a7ff-8612f7d11c21', 'role': 'Model Owner'}
        assert expectedResponse == responses.calls[0].response.json()

    @responses.activate
    def test_client_init_config_param(self):
        self._mock_get_user()

        ArthurAI(config={'access_key': ACCESS_KEY, 'url': BASE_URL})
        assert len(responses.calls) == 1
        expectedResponse = {'organization_id': '65624ee6-9a73-4fe3-a7ff-8612f7d11c21', 'role': 'Model Owner'}
        assert expectedResponse == responses.calls[0].response.json()

    @responses.activate
    def test_client_init_bad_host(self):
        with pytest.raises(UserValueError, match=".*please ensure the URL is correct.*"):
            ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1

    @responses.activate
    def test_client_init_bad_access_key(self):
        self._mock_get_user_unauthorized()
        with pytest.raises(UserValueError, match=".*please ensure your access key is correct.*"):
            ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        assert len(responses.calls) == 1

    @mock.patch('arthurai.client.client.AuthRefresher', MockAuthRefresher)
    @responses.activate
    def test_client_login_with_username_password(self):
        self._mock_user_login()
        self._mock_get_user()

        arthur = ArthurAI(login="arthur", password="password", url=BASE_URL)
        assert USER_TOKEN == arthur.client.session.headers['Authorization']
        assert arthur.verify_ssl is True
        assert arthur.client.session.verify is True

        arthur = ArthurAI(login="arthur", password="password", url=BASE_URL, verify_ssl=False)
        assert arthur.client.session.verify is False

    @responses.activate
    def test_client_init_offline(self):
        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL, offline=True)
        assert len(responses.calls) == 0

    @responses.activate
    def test_client_init_offlines(self):
        mock_get(f"/api/v3/models/1234", status=200,
                 response_body={"organization_id": "65624ee6-9a73-4fe3-a7ff-8612f7d11c21", "role": "Model Owner"})
        ArthurAI(access_key=ACCESS_KEY, url=BASE_URL, offline=True)
        assert len(responses.calls) == 0

    @mock.patch('arthurai.client.client.AuthRefresher', MockAuthRefresher)
    @responses.activate
    def test_client_sets_model_id_and_user_header(self):
        model_data = {
            "id": "1234",
            "partner_model_id": "",
            "input_type": InputType.Tabular,
            "output_type": OutputType.Regression,
            "display_name": "",
            "description": "",
            "attributes": [],
        }
        mock_get("/api/v3/models/1234", status=200, response_body=model_data)

        self._mock_get_user()

        client = ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        client.get_model('1234')

        assert len(responses.calls) == 2
        if "models" in responses.calls[1].request.url:
            header = responses.calls[1].request.headers
        else:
            header = responses.calls[0].request.headers
        assert 'User-Agent' in header
        print(header['User-Agent'])
        assert re.findall(r"^arthur-sdk\/\d.\d+.\d|\w+ \(system=\w*\)$",
                          header['User-Agent'])
        assert os.getenv("ARTHUR_LAST_MODEL_ID") == "1234"

    @responses.activate
    def test_connection_get_model_group(self):
        model_group_id = "01234567-89ab-cdef-0123-456789abcdef"
        model_group_name = "Test Model Group"
        model_group_description = "Test Model Group Description"
        model_group_archived = False
        model_group_created_at = "2022-01-20T23:22:18.185267Z"
        model_group_updated_at = "2022-01-20T23:22:32.914068Z"

        model_group_data = {
            "id": model_group_id,
            "name": model_group_name,
            "description": model_group_description,
            "archived": model_group_archived,
            "created_at": model_group_created_at,
            "updated_at": model_group_updated_at,
            # "versions": MODEL_RESPONSE_JSON
        }

        self._mock_get_user()
        mock_get(f"/api/v3/model_groups/{model_group_id}",
                 status=HTTPStatus.OK,
                 response_body=model_group_data)

        client = ArthurAI(access_key=ACCESS_KEY, url=BASE_URL)
        model_group_from_id = client.get_model_group(model_group_id)

        model = ArthurModel(model_group_id=model_group_id, partner_model_id='pmid', input_type=InputType.Tabular,
                            output_type=OutputType.Multiclass)
        model_group_from_model = client.get_model_group(model)

        assert model_group_from_id == model_group_from_model


class TestGetMultiStatusResponseFailures:
    # fixtures
    SUCCESS_RESULTS = [{"message": "ok", "status": HTTPStatus.OK}, {"message": "created", "status": HTTPStatus.CREATED}]
    USER_FAIL_RESULTS = [{"message": "not found", "status": HTTPStatus.NOT_FOUND}]
    INTERNAL_FAIL_RESULTS = [{"message": "failure", "status": HTTPStatus.INTERNAL_SERVER_ERROR},
                             {"message": "choose", "status": HTTPStatus.MULTIPLE_CHOICES}]
    SUCCESS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0, "total": len(SUCCESS_RESULTS)},
                    "results": SUCCESS_RESULTS}
    FAILURES_BODY = {"counts": {"success": len(SUCCESS_RESULTS),
                                "failure": len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS),
                                "total": len(SUCCESS_RESULTS) + len(INTERNAL_FAIL_RESULTS) + len(USER_FAIL_RESULTS)},
                     "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}
    MISMATCHED_COUNTS_BODY = {"counts": {"success": len(SUCCESS_RESULTS), "failure": 0,
                                         "total": len(SUCCESS_RESULTS)},
                              "results": SUCCESS_RESULTS + USER_FAIL_RESULTS + INTERNAL_FAIL_RESULTS}

    def test_success(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == []
        assert actual_internal_failures == []

    def test_failures(self):
        response = MockResponse(self.FAILURES_BODY, HTTPStatus.MULTI_STATUS)
        actual_user_failures, actual_internal_failures = validate_multistatus_response_and_get_failures(response)
        assert actual_user_failures == self.USER_FAIL_RESULTS
        assert actual_internal_failures == self.INTERNAL_FAIL_RESULTS

    def test_bad_status(self):
        response = MockResponse(self.SUCCESS_BODY, HTTPStatus.OK)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)

    def test_mismatched_counts(self):
        response = MockResponse(self.MISMATCHED_COUNTS_BODY, HTTPStatus.MULTI_STATUS)
        with pytest.raises(ValueError):
            validate_multistatus_response_and_get_failures(response)


# TODO: add testing around allowed exceptions and what gets raised first (e.g. expected 207 but 404 shouldn't raise
#  InternalValueError)
class TestValidateResponse:
    # fixtures
    JSON_BODY = {"status": "somestatus"}
    BYTES_BODY = b"bodybytes"

    def test_server_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.INTERNAL_SERVER_ERROR)
        with pytest.raises(ResponseServerError):
            validate_response_status(response)

    def test_client_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        with pytest.raises(ResponseClientError):
            validate_response_status(response)

    def test_expected_404_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.NOT_FOUND)
        validate_response_status(response, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        with pytest.raises(ResponseRedirectError):
            validate_response_status(response)

    def test_redirect_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.MOVED_PERMANENTLY)
        validate_response_status(response, allow_redirects=True)

    def test_ok_no_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        validate_response_status(response)

    def test_unexpected_ok_error_response(self):
        response = MockResponse(self.JSON_BODY, HTTPStatus.OK)
        with pytest.raises(InternalValueError):
            validate_response_status(response, expected_status_code=HTTPStatus.CREATED)

    def test_server_error(self):
        code = HTTPStatus.INTERNAL_SERVER_ERROR
        with pytest.raises(ResponseServerError):
            validate_response_status(code)

    def test_client_error(self):
        code = HTTPStatus.NOT_FOUND
        with pytest.raises(ResponseClientError):
            validate_response_status(code)

    def test_expected_404_no_error(self):
        code = HTTPStatus.NOT_FOUND
        validate_response_status(code, expected_status_code=HTTPStatus.NOT_FOUND)

    def test_redirect_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        with pytest.raises(ResponseRedirectError):
            validate_response_status(code)

    def test_redirect_no_error(self):
        code = HTTPStatus.MOVED_PERMANENTLY
        validate_response_status(code, allow_redirects=True)

    def test_ok_no_error(self):
        code = HTTPStatus.OK
        validate_response_status(code)

    def test_unexpected_ok_error(self):
        code = HTTPStatus.OK
        with pytest.raises(InternalValueError):
            validate_response_status(code, expected_status_code=HTTPStatus.CREATED)



from datetime import datetime, timezone, timedelta
from typing import Dict, Any, Union, List
from unittest.mock import patch, MagicMock

import numpy as np
import pytest
import pytz

from arthurai.common.exceptions import UserValueError
from arthurai.util import generate_timestamps, normal_random_ints_fixed_sum, format_timestamp, is_valid_datetime_string

MOCK_NOW_VALUE = datetime(2020, 11, 13, tzinfo=pytz.utc)

generate_timestamp_value_cases = [
    ({'total': 8},
     np.array([1, 1, 1, 1, 1, 1, 1, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 6), datetime(2020, 11, 7), datetime(2020, 11, 8), datetime(2020, 11, 9),
       datetime(2020, 11, 10), datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 13))]),
    ({'total': 4, 'duration': "2d"},
     np.array([1, 2, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 12), datetime(2020, 11, 13))]),
    ({'total': 10, 'duration': "3d"},
     np.array([3, 3, 2, 2]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 10), datetime(2020, 11, 10), datetime(2020, 11, 10), datetime(2020, 11, 11),
       datetime(2020, 11, 11), datetime(2020, 11, 11), datetime(2020, 11, 12), datetime(2020, 11, 12),
       datetime(2020, 11, 13), datetime(2020, 11, 13))]),
    ({'total': 1, 'duration': "3d"}, None, UserValueError),
    ({'total': 11, 'duration': "3h", 'freq': "H"},
     np.array([3, 4, 2, 2]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 11, 12, hour=21), datetime(2020, 11, 12, hour=21), datetime(2020, 11, 12, hour=21),
      datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=22),
      datetime(2020, 11, 12, hour=22), datetime(2020, 11, 12, hour=23), datetime(2020, 11, 12, hour=23),
      datetime(2020, 11, 13), datetime(2020, 11, 13))]),
    ({'total': 4, 'duration': "2d", 'end': datetime(2020, 12, 13, tzinfo=pytz.utc)},
     np.array([2, 1, 1]),
     [pytz.utc.localize(dt) for dt in
      (datetime(2020, 12, 11), datetime(2020, 12, 11), datetime(2020, 12, 12), datetime(2020, 12, 13))]),
]


@pytest.mark.parametrize("kwargs,repeats,expected", generate_timestamp_value_cases)
def test_generate_timestamps_values(kwargs: Dict[str, Any], repeats: np.ndarray, expected: Union[List[datetime], type]):
    """
    Test that generate_timestamps() outputs match expected values for a few cases
    """
    mock_now = MagicMock()
    mock_now.return_value = MOCK_NOW_VALUE
    mock_randints = MagicMock(normal_random_ints_fixed_sum)
    mock_randints.return_value = repeats
    with patch("arthurai.util._datetime_now", mock_now):
        with patch("arthurai.util.normal_random_ints_fixed_sum", mock_randints):
            if isinstance(expected, list):
                actual = generate_timestamps(**kwargs)
                assert actual == expected
            elif issubclass(expected, Exception):
                with pytest.raises(expected):
                    generate_timestamps(**kwargs)
            else:
                raise TypeError("did not understand 'expected' value")


random_ints_cases = [
    {'num_values': 47, 'total_sum': 58730},
    {'num_values': 47, 'total_sum': 47},
    {'num_values': 13, 'total_sum': 2302},
    {'num_values': 92, 'total_sum': 50238}
]


@pytest.mark.parametrize("kwargs", random_ints_cases)
def test_normal_random_ints_fixed_sum(kwargs: Dict[str, Any]):
    np.random.seed(91)
    actual = normal_random_ints_fixed_sum(**kwargs)
    assert len(actual) == kwargs['num_values']
    assert actual.sum() == kwargs['total_sum']


def test_format_timestamp_with_no_tz_parameter():
    """Checks to make sure the handling of datetime objects/string are handled correctly"""
    timestamps = [
        datetime(2020, 8, 13, 17, 44, 31, tzinfo=timezone.utc),
        datetime(2020, 8, 13, 17, 44, 31),
        "2020-08-13T17:44:31Z",
        "2020-08-13T17:44:31",
        20200806174431,
    ]
    correct_outputs = [
        "2020-08-13T17:44:31+00:00",
        ValueError,
        "2020-08-13T17:44:31Z",
        ValueError,
        TypeError
    ]

    for count, timestamp in enumerate(timestamps):
        if correct_outputs[count] == ValueError or correct_outputs[count] == TypeError:
            with pytest.raises(correct_outputs[count]):
                format_timestamp(timestamp)
        else:
            assert format_timestamp(timestamp) == correct_outputs[count]


def test_format_timestamps_with_tz_parameter():
    """Checks to make sure the handing of datetime objects/string with timezone info"""
    timestamps = [
        (datetime(2020, 8, 13, 17, 44, 31), 'US/Eastern'),
        (datetime(2020, 8, 13, 17, 44, 31), timezone(timedelta(hours=1)))
    ]

    correct_outputs = [
        '2020-08-13T21:44:31+00:00',
        '2020-08-13T16:44:31+00:00'
    ]

    for count, t in enumerate(timestamps):
        assert format_timestamp(t[0], t[1]) == correct_outputs[count]


def test_is_valid_datetime_string():
    """Checks a few edge cases for the is_valid_datetime_string"""
    input_dates = [
        '2020-08-13T21:44:31+00:00',
        "2020-08-13T17:44:31Z",
        2020081321443514,
        [1234, 51234, 12351],
        9.1,
    ]

    correct_outputs = [
        True,
        True,
        False,
        False,
        False
    ]

    for count, date in enumerate(input_dates):
        assert is_valid_datetime_string(date) == correct_outputs[count]


import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import Any, List, Union

from arthurai.common.exceptions import arthur_excepted
from arthurai.core.viz import style, utils
import warnings
warnings.filterwarnings('ignore')


class DataVisualizer(object):
    def __init__(self, arthur_models):
        self.models = arthur_models

    @arthur_excepted("failed to generate timeline")
    def timeline(self, attribute_name):
        """Generates a visualization of the distribution of an attribute over time. 
        For categorical attributes, a stacked area chart over time.
        For continuous attributes, a joyplot showing probability densities over time.
        :raises ArthurUserError: failed due to user error
        :raises ArthurInternalError: failed due to an internal error
        """
        for model in self.models:
            if model.get_attribute(attribute_name).categorical:
                self._timeline_categorical(model, attribute_name)
            else:
                self._timeline_continuous(model, attribute_name)

    @arthur_excepted("failed to generate metric series")
    def metric_series(self, metric_names, time_resolution="day"):
        """Generates a line series visualization for selected metrics.

        .. code-block:: python

            # plot both the model's AUC metric and the model's FPR metric by the hour
            arthur_model.viz.metric_series(["auc", "falsePositiveRate"], time_resolution="hour")

        :raises ArthurUserError: failed due to user error
        :raises ArthurInternalError: failed due to an internal error
        """
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)
        predicted_property, ground_truth_property = utils.get_pred_and_gt_attrs(self.models[0])
        pal = style.categorical_palette

        # Define list of palettes to be used
        num_metrics = len(metric_names)
        max_metrics = len(style.multi_categorical_palette_names)
        if num_metrics > max_metrics:
            raise UserValueError(f"Too many metrics to display. The maximum is {max_metrics}, you wanted to see {num_metrics}")
        color_palette_names = style.multi_categorical_palette_names[0:num_metrics]

        for metric_index, metric_name in enumerate(metric_names):
            color_palette = sns.color_palette(palette=color_palette_names[metric_index], n_colors=len(self.models))
            for model_index, model in enumerate(self.models):
                if model.is_batch:
                    time_resolution = "batch_id"
                    query = _metric_series_batch_query(metric_name, ground_truth_property,
                        predicted_property, model.classifier_threshold)
                else:
                    query = _metric_series_streaming_query(metric_name, time_resolution,
                        ground_truth_property, predicted_property, model.classifier_threshold)

                response = model.query(query)
                df = pd.DataFrame(response)[::-1]

                label = _label(model, metric_name)
                color = _alpha_color(color_palette[model_index])
                plt.plot(df[time_resolution], df[metric_name], lw=12, color=color, label=label)
                plt.plot(df[time_resolution], df[metric_name], lw=2, color=color)
                plt.xticks([])
                ax.spines['right'].set_visible(False)
                ax.spines['top'].set_visible(False)
                plt.legend(loc="upper left", facecolor="white")

    @arthur_excepted("failed to generate drift series")
    def drift_series(self, attribute_names: Union[str, List[str]], drift_metric: str = "PSI", time_resolution: str = "day"):
        """Generates a line series visualization of data drift metrics for selected attributes.

        .. code-block:: python

            # plot the KLDivergence drift of features X1, X2, and X3 by the hour
            arthur_model.viz.drift_series(["X1", "X2", "X3"], drift_metric="KLDivergence", time_resolution="hour")

        :raises ArthurUserError: failed due to user error
        :raises ArthurInternalError: failed due to an internal error
        """
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)
        labels = []

        # if attribute_names is a string, convert to list with one element
        if isinstance(attribute_names, str):
            attribute_names = [attribute_names]

        # Define list of palettes to be used
        color_palettes = _drift_color_palettes(len(attribute_names), len(self.models))

        for model_index, model in enumerate(self.models):
            if model.is_batch:
                time_resolution = "batch_id"
            timestamp_query = _batch_timestamp_query()
            drift_query = {
                "properties": attribute_names,
                "num_bins": 20,
                "rollup": time_resolution,
                "base": {
                "source": "reference"
                },
                "target": {
                "source": "inference"
                },
                "metric": drift_metric
            }
            if model.is_batch:
                drift_df = pd.DataFrame(model.query(drift_query, query_type="drift"))
                timestamp_df = pd.DataFrame(model.query(timestamp_query))
                df = drift_df.rename(columns={"rollup":time_resolution}
                    ).join(timestamp_df.set_index(time_resolution), on=time_resolution
                    ).sort_values(by="timestamp"
                    ).set_index(time_resolution)
            else:
                df = pd.DataFrame(model.query(drift_query, query_type="drift")
                        ).sort_values(by="rollup").set_index("rollup")

            for attribute_name in attribute_names:
                labels.append(_label(model, attribute_name))

            # base_color_palette = style.categorical_palette
            color_palette = color_palettes[model_index]
            # This darkens the color for each successive model version. There are probably better way of doing this
            # but this is just a hack to get it to work for now.
            # for color_tuple in base_color_palette:
            #     color_palette.append((color_tuple[0] / (2 ** model_index), color_tuple[1] / (2 ** model_index), color_tuple[2] / (2 ** model_index), color_tuple[3]))
            df[attribute_names].plot(ax=ax, color=color_palette, lw=8)
            plt.ylabel(drift_metric)
            plt.xticks([])
            plt.xlabel("")
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)
        plt.legend(labels, loc="upper left", facecolor="white")

    def _timeline_categorical(self, model, attribute_name, time_resolution="day"):
        """
        """
        print(_label(model, attribute_name))
        if model.is_batch:
            time_resolution="timestamp"
            query = _timeline_categorical_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_categorical_streaming_query(attribute_name, time_resolution)
        response = model.query(query)
        df = pd.DataFrame(response
            ).sort_values(by=time_resolution
            ).pivot(index=time_resolution,columns=attribute_name, values="count"
            ).fillna(0.0)

        sns.set(style="white")
        fig = plt.figure(figsize=(20,4))
        ax = fig.add_subplot(1, 1, 1)
        plt.stackplot(df.index.values, df.values.T.tolist(),
            labels=df.columns.values, colors=style.categorical_palette)
        plt.xticks([])
        plt.ylabel("Count")
        plt.legend(loc='upper right')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        plt.show()



    def _timeline_continuous(self, model, attribute_name, time_resolution="day"):
        """
        """
        print(_label(model, attribute_name))
        if model.is_batch:
            time_resolution="timestamp"
            query = _timeline_continuous_batch_query(attribute_name, time_resolution)
        else:
            query = _timeline_continuous_streaming_query(attribute_name, time_resolution)
        response = model.query(query)

        dfs= []
        for group in response:
            temp_df= pd.DataFrame(group["distribution"])
            temp_df[time_resolution] = group[time_resolution]
            temp_df["count"] = utils.savgol_filter(temp_df["count"])
            dfs.append( temp_df)
        df = pd.concat(dfs).sort_values([time_resolution, "lower"])

        global_y_min = df["lower"].min()
        global_y_max = df["lower"].max()
        global_x_min = df["count"].min()
        global_x_max = df["count"].max()

        sns.set(style="white", rc={"axes.facecolor": (0, 0, 0, 0)})
        # Initialize the FacetGrid object
        pal = style.continuous_palette
        g = sns.FacetGrid(df, col=time_resolution, hue=time_resolution, aspect=.2, height=8, palette=pal, subplot_kws={"xlim":(global_x_max, global_x_min), "ylim":(global_y_min, global_y_max)})

        # Draw the densities in a few steps
        g.map(plt.plot, "count", "lower",clip_on=False,  color="white", lw=4.5,  alpha=0.6)
        g.map(plt.fill_between, "count","lower", clip_on=False, lw=2)

        # Set the subplots to overlap
        g.fig.subplots_adjust(wspace=-.55, left = 0.125)
        g.set_titles("")
        g.set(yticks=[])
        g.set(xticks=[])
        g.set(ylabel="")
        g.set(xlabel="")
        g.despine(bottom=True, left=True)


def _metric_series_streaming_query(metric_name, time_resolution, ground_truth_property, predicted_property, classifier_threshold):
    """Generates query for fetching metrics for streaming model.
    """
    query = {
                "select": [
                    {
                        "function": metric_name,
                        "alias": metric_name,
                        "parameters": {
                            "ground_truth_property": ground_truth_property,
                            "predicted_property": predicted_property,
                            "threshold": classifier_threshold
                        }

                    },

                    {
                        "function": "roundTimestamp",
                        "alias": time_resolution,
                        "parameters": {
                            "property": "inference_timestamp",
                            "time_interval": time_resolution
                        }
                    }
                ],
                "group_by": [
                    {
                        "alias": time_resolution
                    }
                ],
                "order_by": [
                    {
                        "alias": time_resolution,
                        "direction": "desc"
                    }
                ]
            }
    return query


def _metric_series_batch_query(metric_name, ground_truth_property, predicted_property, classifier_threshold):
    """Generates query for fetching metrics for batch model.
    """
    query = {
                "select": [
                    {"function": metric_name,
                        "alias": metric_name,
                        "parameters": {
                            "ground_truth_property": ground_truth_property,
                            "predicted_property": predicted_property,
                            "threshold": classifier_threshold
                        }

                    },
                    {
                        "function":"max",
                        "parameters": {"property":"inference_timestamp"},
                        "alias":"timestamp"
                    },
                    {"property":"batch_id"}
                ],
                "group_by": [
                    {
                        "property":"batch_id"
                    }
                ],
                "order_by": [
                    {
                        "alias":"timestamp",
                        "direction": "desc"
                    }
                ]
            }
    return query


def _timeline_continuous_streaming_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a streaming model.
    """
    return {
        "select":[
            {
                "function":"distribution",
                "alias":"distribution",
                "parameters":{"property": attribute_name, "num_bins":50}
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution
                    }
            }
        ],
        "group_by": [{"alias":time_resolution} ],
        "order_by": [{"alias":time_resolution, "direction": "desc"} ]
    }


def _timeline_continuous_batch_query(attribute_name, time_resolution):
    """Generates a query for continuous attribute in a batch model.
    """
    return {
        "select":[
            {
                "function":"distribution",
                "alias":"distribution",
                "parameters":{"property": attribute_name, "num_bins":50}
            },
            {
                "function":"max",
                "alias":"timestamp",
                "parameters": {"property":"inference_timestamp"}
            },
            {"property": "batch_id"}
        ],
        "group_by": [
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }


def _timeline_categorical_streaming_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a streaming model.
    """
    return {
        "select": [
            {
                "property": attribute_name
            },
            {
                "function": "count"
            },
            {
                "function": "roundTimestamp",
                "alias": time_resolution,
                "parameters": {
                    "property": "inference_timestamp",
                    "time_interval": time_resolution
                }
            }
        ],
        "group_by": [
            {
                "property": attribute_name
            },
            {
                "alias": time_resolution
            }
        ],
        "order_by": [
            {
                "property": attribute_name
            },
            {
                "alias": time_resolution,
                "direction": "desc"
            }
        ]
    }


def _timeline_categorical_batch_query(attribute_name, time_resolution):
    """Generates a query for categorical attribute in a batch model.
    """
    return {
        "select": [
            {
                "property": attribute_name
            },
            {
                "function": "count"
            },
            {
                "property":"batch_id"
            },
            {"function":"max", "parameters": {"property":"inference_timestamp"}, "alias":"timestamp"}
        ],
        "group_by": [
            {
                "property": attribute_name
            },
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "property": attribute_name
            },
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }


def _batch_timestamp_query():
    return {
        "select":[
            {
                "function":"max",
                "alias":"timestamp",
                "parameters": {"property":"inference_timestamp"}
            },
            {"property": "batch_id"}
        ],
        "group_by": [
            {
                "property":"batch_id"
            }
        ],
        "order_by": [
            {
                "alias":"timestamp",
                "direction": "desc"
            }
        ]
    }


def _label(model, label):
    model_name = model.display_name
    if len(model_name) <= 20:
        short_model_name = model_name
    else:
        short_model_name = model_name[0:17] + "..."
    return "(" + str(model.version_sequence_num) + ")__" + short_model_name + "__" + label


def _alpha_color(color):
    return (color[0], color[1], color[2], style.alpha)


def _drift_color_palettes(num_attributes, num_models):
    max_attributes = len(style.multi_categorical_palette_names)
    if num_attributes > max_attributes:
        raise UserValueError(
            f"Too many attributes to display. The maximum is {max_attributes}, you wanted to see {num_attributes}")

    color_palettes = []
    color_palette_names = style.multi_categorical_palette_names[0:num_attributes]
    for i in range(num_models):
        color_palettes.append([])
    for color_palette_name in color_palette_names:
        base_palette = sns.color_palette(palette=color_palette_name, n_colors=num_models)
        for i, base_palette_color in enumerate(base_palette):
            color_palettes[i].append(_alpha_color(base_palette_color))

    return color_palettes



from typing import Any, List, Optional
from dataclasses import dataclass

from arthurai.core.base import ArthurBaseJsonDataclass

@dataclass
class AuthInfo(ArthurBaseJsonDataclass):
    issuer: str
    username: str
    first_name: str
    last_name: str
    email: str
    roles: List[str]
    organization_ids: List[str]
    internal_user_id: Optional[str] = None
    external_user_id: Optional[str] = None

    def __post_init__(self):
        """
        Special initialization method for dataclasses that is called after the generated __init__() method.

        Input parameters to __post_init__ (may) be parsed out of the class variables and into this
        method. E.g. defining ArthurModel.client allows you to create an ArthurModel instance as
        `ArthurModel(client=...)` where client is only passed into __post_init__ and does not show
        up as an instance variable. To do so, the class variable type must be defined with an
        InitVar[] wrapper (refer to link to Python docs below).
        https://docs.python.org/3/library/dataclasses.html#init-only-variables
        
        Variables created here will only be accessible directly on the object itself, they will not
        be in the result of object.to_dict() even if marked as public (does not have preceding
        underscore).
        """
        pass




API_PREFIX = "/api/v3"


class ListableStrEnum:
    """
    This class operates similar to Enum but passes mypy type checking.
    The members are accessed directly rather than by `.name` or `.value`.
    """

    @classmethod
    def list(cls):
        """
        Lists all attributes in alphabetical order
        """
        members = [getattr(cls, attr) for attr in dir(cls)
                   if not callable(getattr(cls, attr)) and not attr.startswith("__")]
        return sorted(members)


class InputType(ListableStrEnum):
    Tabular = "TABULAR"
    Image = "IMAGE"
    NLP = "NLP"


class OutputType(ListableStrEnum):
    Regression = "REGRESSION"
    Multiclass = "MULTICLASS"
    Multilabel = "MULTILABEL"
    ObjectDetection = "OBJECT_DETECTION"


class ValueType(ListableStrEnum):
    String = "STRING"
    Integer = "INTEGER"
    Float = "FLOAT"
    Image = "IMAGE"
    Boolean = "BOOLEAN"
    Timestamp = "TIMESTAMP"
    Unstructured_Text = "UNSTRUCTURED_TEXT"  # don't remove old one, backward compatibility
    BoundingBox = "BOUNDING_BOX"


class Stage(ListableStrEnum):
    ModelPipelineInput = "PIPELINE_INPUT"
    PredictFunctionInput = "PREDICT_FUNCTION_INPUT"
    """PREDICT_FUNCTION_INPUT is deprecated in 3.21.1. Use PIPELINE_INPUT instead."""
    PredictedValue = "PREDICTED_VALUE"
    NonInputData = "NON_INPUT_DATA"
    GroundTruth = "GROUND_TRUTH"
    GroundTruthClass = "GROUND_TRUTH_CLASS"


class TextDelimiter(ListableStrEnum):
    """This class contains patterns that can be used as text_delimiter for NLP models.
    """
    NOT_WORD = "\W+"
    """``"\W+"`` Splits on any character that is not a word.
    
    Ex: ``"this  is,aaaa,,,,,test!" = ["this", "is", "aaaa", "test", ""]``
    """

    WHITESPACE = "\s+"
    """``"\s+"`` Splits on whitespace.
    
    Ex: ``"this  is,a test! " = ["this", "is,a", "test!", ""]``"""

    COMMA = ","
    """``","`` Splits on a single comma.
    
    Ex: ``"this ,is,,a,test" = ["this ", "is", "", "a", "test"]``"""

    COMMA_PLUS = ",+"
    """``",+"`` Splits on one or more commas.
    
    Ex: ``"this ,is,,a,test" = ["this ", "is", "a", "test"]``"""

    PIPE = "\|"
    """``"\|"`` Splits on a single pipe.
    
    Ex: ``"this |is||a|test" = ["this ", "is", "", "a", "test"]``"""

    PIPE_PLUS = "\|+"
    """``"\|+"`` Splits on one or more pipes.
    
    Ex: ``"this |is||a|test" = ["this ", "is", "a", "test"]``"""


class Enrichment(ListableStrEnum):
    """This class contains constants for the names of enrichments"""
    AnomalyDetection = "anomaly_detection"
    BiasMitigation = "bias_mitigation"
    Explainability = "explainability"
    Hotspots = "hotspots"


class ImageResponseType(ListableStrEnum):
    """Valid image response types supported by Arthur API"""
    RawImage = 'raw_image'
    ResizedImage = 'resized_image'
    Thumbnail = 'thumbnail'
    LimeExplanation = 'lime_explanation'


class ImageContentType(ListableStrEnum):
    """Valid image content types supported by Arthur API"""
    Png = 'image/png'
    Jpeg = 'image/jpeg'
    Gif = 'image/gif'
    Tiff = 'image/tiff'


IMAGE_FILE_EXTENSION_MAP = {
    ImageContentType.Png: '.png',
    ImageContentType.Jpeg: '.jpg',
    ImageContentType.Gif: '.gif',
    ImageContentType.Tiff: '.tff'
}


class AccuracyMetric(ListableStrEnum):
    Accuracy = "accuracy"
    Recall = "recall"
    F1 = "f1"
    Precision = "precision"


class Role(ListableStrEnum):
    User = "User"
    ModelOwner = "Model Owner"
    Administrator = "Administrator"


class ModelStatus(ListableStrEnum):
    Ready = "Ready"
    CreationFailed = "CreationFailed"
    Creating = "Creating"
    Pending = "Pending"
    Archived = "Archived"
    Archiving = "Archiving"
    ArchiveFailed = "ArchiveFailed"
    Unknown = "Unknown"


ONBOARDING_SPINNER_MESSAGE = "Waiting for the model to be ready to accept inferences."
ONBOARDING_UPDATE_MESSAGE = "We are still working on getting your model ready to accept inferences..."
DEFAULT_SERVICE_ACCOUNT = "default_sdk_service_account"




from typing import Optional
from arthurai.common.constants import Stage, OutputType

def check_has_bias_attrs(model):
    """
    Returns True if and only if the model has any attributes being monitored for bias by Arthur
    """
    for attr in model.attributes:
        if attr.monitor_for_bias:
            return True 
    return False

def check_attr_is_bias(model, attr_name: str):
    """
    Returns True if and only if the model has an attribute with the given name `attr_name` being monitored for bias by
    Arthur
    """
    attr = model.get_attribute(attr_name)
    return attr.monitor_for_bias

def get_positive_predicted_class(model):
    """
    Checks if model is a binary classifier. Returns False if multiclass, otherwise returns the name of the positive
    predicted attribute
    """
    if model.output_type != OutputType.Multiclass:
        return("Not a classifier model.")
    predicted_value_attributes = model.get_attributes(stage=Stage.PredictedValue)
    if len(predicted_value_attributes) != 2:
        return ("Not a binary classifier.")
    for attr in predicted_value_attributes:
        if attr.is_positive_predicted_attribute:
            return attr.name



from abc import ABCMeta, abstractmethod


class AbstractHTTPClient(metaclass=ABCMeta):

    @abstractmethod
    def send(self, endpoint, method, json=None, files=None, headers=None, params=None, **kwargs):
        pass

    @abstractmethod
    def get(self, endpoint, headers=None, params=None, **kwargs):
        pass

    @abstractmethod
    def post(self, endpoint, json=None, files=None, headers=None, **kwargs):
        pass

    @abstractmethod
    def patch(self, endpoint, json=None, files=None, headers=None, **kwargs):
        pass

    @abstractmethod
    def put(self, endpoint, json=None, files=None, headers=None, **kwargs):
        pass

    @abstractmethod
    def delete(self, endpoint, headers=None, **kwargs):
        pass




# download.py
import boto3
from botocore import UNSIGNED, exceptions
from botocore.client import Config
from enum import Enum
import os.path
from os import path
import logging
from pathlib import Path
from typing import Dict, List
import zipfile
import pandas as pd
import yaml

from arthurai.common.exceptions import ArthurDownloadError

logger = logging.getLogger(__name__)

s3_bucket_name = "s3-bucket-arthur-public"
default_download_directory = Path("arthur-downloads/")


class ArthurDatasetSource(Enum):
    """
    Class to enumerate the sources of data that can be used for an ArthurExampleDownloader
    """
    S3 = "S3"


class ArthurExampleDownloader:
    """
    Class for downloading data for an Arthur Example
    """

    def __init__(self, example_name: str, target_base_dir: Path, source=ArthurDatasetSource.S3):
        self.example_name = example_name
        self.source = source
        if not os.path.isdir(target_base_dir):
            os.mkdir(target_base_dir)
        if not os.path.isdir(target_base_dir / example_name):
            os.mkdir(target_base_dir / example_name)
        self.target_path = target_base_dir / example_name

        # scan each sub-folder within sandbox/example_name/ :
        # metadata/, data/, predictions/, and models/
        if self.source == ArthurDatasetSource.S3:
            self.metadata_base_path = Path(f"sandbox/{self.example_name}/metadata/")
            self.dataset_base_path = Path(f"sandbox/{self.example_name}/data/")
            self.pred_base_path = Path(f"sandbox/{self.example_name}/predictions/")
            self.model_base_path = Path(f"sandbox/{self.example_name}/models/")

            metadata_files = get_file_keys_in_s3_folder(self.metadata_base_path)
            if len(metadata_files) == 1:
                self.metadata_file = metadata_files[0]
            else:
                raise ArthurDownloadError("Currently can only support downloading 1 metadata file from S3 (not 0 or >=2)")

            dataset_files = get_file_keys_in_s3_folder(self.dataset_base_path)
            if len(dataset_files) == 1:
                self.dataset_file = dataset_files[0]
            else:
                raise ArthurDownloadError("Currently can only support downloading 1 dataset file from S3 (not 0 or >=2)")

            pred_files = get_file_keys_in_s3_folder(self.pred_base_path)
            if len(pred_files) == 1:
                self.pred_file = pred_files[0]
            else:
                raise ArthurDownloadError("Currently can only support downloading 1 predictions file from S3 (not 0 or >=2)")

            model_files = get_file_keys_in_s3_folder(self.model_base_path)
            if len(model_files) == 0:
                self.model_file = None
            elif len(model_files) == 1:
                self.model_file = model_files[0]
            else:
                raise ArthurDownloadError("Currently can only support downloading <=1 model file from S3 (not >=2)")
        else:
            raise ValueError(f"Currently not able to support downloading from {source}")

    def download_metadata(self) -> Dict:
        """Gets metadata for an ArthurExample

        :return: metadata dictionary with parameters relating to file locations & ArthurModel model & attribute info
        """
        if self.source == ArthurDatasetSource.S3:
            try:
                s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
                s3_filepath = self.metadata_base_path / self.metadata_file
                response = s3_client.get_object(Bucket=s3_bucket_name, Key=str(s3_filepath))
                metadata = yaml.safe_load(response["Body"])
                return metadata
            except exceptions.ClientError:
                raise ArthurDownloadError(f"Could not get {self.example_name} metadata from S3. Check if a) you have "
                                          f"spelled the name correctly or b) {self.example_name} is in fact an example "
                                          f"with data in the Arthur public S3 bucket in the sandbox folder.")
        else:
            raise ValueError(f"Currently not able to support downloading from {self.source}")

    def download_dataset(self) -> Path:
        """Gets an example dataset for an ArthurExample

        :return: Path to the downloaded dataset
        """

        if not path.exists(self.target_path / self.dataset_file):
            download_from_s3(self.dataset_base_path / self.dataset_file, self.target_path / self.dataset_file)
        return self.target_path / self.dataset_file

    def download_pretrained_model(self) -> Path:
        """Gets a pretrained model for an ArthurExample

        :return: Path to the downloaded model
        """
        if self.model_file is not None:
            if not path.exists(self.target_path / self.model_file):
                download_from_s3(self.model_base_path / self.model_file, self.target_path / self.model_file)
            return self.target_path / self.model_file
        raise ValueError("There is no model to be downloaded for this example")

    def download_pretrained_model_predictions(self) -> Path:
        """Gets pretrained model predictions for an ArthurExample

        :return: Path to the downloaded predictions
        """
        if not path.exists(self.target_path / self.pred_file):
            download_from_s3(self.pred_base_path / self.pred_file, self.target_path / self.pred_file)
        return self.target_path / self.pred_file


def download_from_s3(s3_file_address: Path, local_file_destination: Path) -> None:
    """Downloads a file from S3 to be saved locally (and unzips if it is a zipped file)

    :param s3_file_address: Path, the full location filepath for the file in the Arthur public S3 bucket
    :param local_file_destination: Path, the directory to save the file to locally
    :return: None
    """
    s3_client = boto3.resource('s3', config=Config(signature_version=UNSIGNED))
    try:
        local_file_destination_string = str(local_file_destination)
        s3_client.Object(str(s3_bucket_name), str(s3_file_address)).download_file(local_file_destination_string)
        if "zip" in local_file_destination_string:
            with zipfile.ZipFile(local_file_destination, 'r') as zip_ref:
                zip_ref.extractall()
    except exceptions.ClientError:
        raise ArthurDownloadError(f"Could not download {s3_file_address} from S3. Check if a) you have spelled the "
                                  f"filename correctly or b) {s3_file_address} is actually hosted on the Arthur public "
                                  f"S3 bucket in the sandbox folder.")


def get_file_keys_in_s3_folder(filepath: Path) -> List[str]:
    """
    Lists the files in the S3 directory

    :param filepath: Path, the path of the S3 directory to read files from
    """
    s3_client = boto3.resource('s3', config=Config(signature_version=UNSIGNED))
    bucket = s3_client.Bucket(s3_bucket_name)
    return [o.key.split('/')[-1] for o in bucket.objects.filter(Prefix=str(filepath)) if o.key[-1] != "/"]


def load_downloaded_file(filepath: Path) -> pd.DataFrame:
    """Returns a file located at `filepath` to the user

    :param filepath: Path, the local path to load the file from
    """
    filepath_string = str(filepath)
    if "csv" in filepath_string:
        return pd.read_csv(filepath_string)
    elif "parquet" in filepath_string:
        return pd.read_parquet(filepath_string)
    else:
        raise TypeError(f"Currently not able to support returning the filetype in {filepath_string}")




from base64 import b64encode, b64decode
from datetime import datetime, timedelta
from typing import Dict, Tuple

import jwt
import pytz

from arthurai.client.helper import user_login
from arthurai.common.constants import API_PREFIX


class AuthRefresher:

    AUTH_KEY = 'Authorization'
    ALGORITHMS = ['HS256']
    MINS_BEFORE_EXPIRY_TO_REFRESH = 5

    def     __init__(self, url: str, login: str, password: str, verify_ssl: bool):
        self._url = url
        self._login = login
        # encode password to obfuscate a bit
        self._password_encoded = b64encode(password.encode())
        self._verify_ssl = verify_ssl

    @staticmethod
    def _get_refresh_wait_time(current_token: str) -> timedelta:
        """
        Given a JWT token with an 'exp' field, return how long to wait before refreshing the token
        :param current_token: the current token being used, containing an 'exp' field
        :return: the amount of time to wait before fetching a new token
        """
        # see when the token expires
        token_decoded = jwt.decode(current_token, algorithms=AuthRefresher.ALGORITHMS,
                                   options={"verify_signature": False})
        expiry = datetime.fromtimestamp(token_decoded['exp'], tz=pytz.UTC)
        cur_time = datetime.now(tz=pytz.UTC)

        time_to_briefly_before_expiry = (expiry -
                                         timedelta(minutes=AuthRefresher.MINS_BEFORE_EXPIRY_TO_REFRESH)) - cur_time
        # ensure time isn't negative
        if time_to_briefly_before_expiry < timedelta():
            return timedelta()
        else:
            return time_to_briefly_before_expiry

    def refresh(self) -> Tuple[Dict[str, str], timedelta]:
        """
        Authorization header update function for an HTTPClient

        Fetches a new session token and returns the new token, and how long to wait before refreshing
        it (by calling this method again)
        :return: Headers to update (Authorization), and time to wait before refreshing again
        """
        password = b64decode(self._password_encoded).decode()
        auth_token, _ = user_login(api_http_host=self._url, api_prefix=API_PREFIX, login=self._login,
                                   password=password, verify_ssl=self._verify_ssl)
        next_refresh_wait_time = self._get_refresh_wait_time(auth_token)
        return {self.AUTH_KEY: auth_token}, next_refresh_wait_time



from typing import List, Union, Optional

from dataclasses import dataclass

from arthurai.common.constants import ValueType, Stage
from arthurai.common.exceptions import MissingParameterError, UserValueError
from arthurai.core.base import ArthurBaseJsonDataclass


MAX_CATEGORIES_TO_SHOW = 5


def get_attribute_order_stage(attribute_stage: Stage) -> int:
    if attribute_stage == Stage.ModelPipelineInput:
        return 0
    elif attribute_stage == Stage.NonInputData:
        return 1
    elif attribute_stage == Stage.PredictedValue:
        return 2
    elif attribute_stage == Stage.GroundTruth:
        return 3
    elif attribute_stage == Stage.GroundTruthClass:
        return 4
    else:
        return 5


@dataclass
class AttributeCategory(ArthurBaseJsonDataclass):
    """
    A list of the attribute's categories. An attribute will only have categories if it is marked as categorical.
    """
    value: str
    label: Optional[str] = None

    def __post_init__(self):
        self.value = str(self.value)
        if self.label is not None:
            self.label = str(self.label)

    def __str__(self):
        return str(self.value)


@dataclass
class AttributeBin(ArthurBaseJsonDataclass):
    """
    A list of the attribute's bins. An attribute will only have bins if it is not categorical. The bin start is
     exclusive and the end is inclusive, (continuous_start, continuous_end]. Use Null to represent an open end of a bin.
    """
    continuous_start: Optional[float] = None
    continuous_end: Optional[float] = None


@dataclass
class ArthurAttribute(ArthurBaseJsonDataclass):
    """ArthurAttribute encapsulates data associated with a model attribute

    :param attribute_link: Only applicable for `GroundTruth` or `PredictedValue` staged attributes.
        If stage is equal to `GroundTruth`, this represents the associated `PredictedValue` attribute and vice versa
    :param is_positive_predicted_attribute: Only applicable for `PredictedValue` attributes on a Binary
        Classification model. Should be set to `True` on the positive predicted value attribute.
    :param is_unique: Boolean value used to signal if the values of this attribute are unique. To set an attribute to be
        unique, use :code:`arthur_model.get(name=attribute_name).set(is_unique=True)`.
    :param bins: List of bin cut-offs used to discretize continuous attributes. Use `None` as an open ended value.
        ``[None, 18, 65, None]`` represents the three following bins: ``value < 18, 18 < value < 65, value > 65``.
        To modify an attribute's bins, use :code:`arthur_model.get_attribute(name=attribute_name).set(bins=[None, cutoff_1, ..., cutoff_n, None])`.
    :param monitor_for_bias: boolean value set to `True` if the attribute should be monitored for bias. To modify an attributes's
        bias monitoring status, use :func:`arthurai.core.models.ArthurModel.set_attribute_as_sensitive()`.
    :param max_range: Max value for a continuous attribute. To modify an attribute's range, use :code:`arthur_model.get(name=attribute_name).set(range=[min_value, max_value])`.
    :param min_range: Min value for a continuous attribute. To modify an attribute's range, use :code:`arthur_model.get(name=attribute_name).set(range=[min_value, max_value])`.
    :param categorical: Boolean value set to `True` if the attribute has categorical values. To set an attribute to be
        categorical, use :code:`arthur_model.get_attribute(name=attribute_name).set(categorical=true)`.
    :param position: The array position of attribute within the stage. Required in the PREDICT_FUNCTION_INPUT stage.
    :param label: Label for attribute. If attribute has an encoded name, a more readable label can be set.
    :param stage: :class:`arthurai.common.constants.Stage` of this attribute in the model pipeline. To modify an attribute's
        stage, use :code:`arthur_model.get(name=attribute_name).set(stage=stage_value)`. See :class:`arthurai.common.constants.Stage`
        for the different stage values an attribute can have.
    :param value_type: :class:`arthurai.common.constants.ValueType` associated with this attributes values. Non-modifiable.
    :param name: Name of the attribute. Attribute names can only contain alpha-numeric characters and underscores
        and cannot start with a number. To modify an attribute's name, use :func:`arthurai.core.models.ArthurModel.rename_attribute()`.
    :param categories: [Only for Categorical Attributes] If the attribute is categorical, this will contain the
        attribute's categorical values and the labels for each value. It is required only if the categorical flag
        is set to true. To modify the category values, use :code:`arthur_model.get(name=attribute_name).set([{'value': value_1}, ..., {'value': value_n}])`.
        To mofidy the category labels, use :func:`arthurai.core.models.ArthurModel.set_attribute_labels()`.
    :param gt_class_link: Optional link for a predicted attribute to its corresponding value
        in a ground truth class attribute
    :param implicit: is the attribute implicitly created by the API
    """

    name: str
    value_type: ValueType
    stage: Stage
    id: Optional[str] = None
    label: Optional[str] = None
    position: Optional[int] = None
    categorical: Optional[bool] = False
    min_range: Optional[Union[int, float]] = None
    max_range: Optional[Union[int, float]] = None
    monitor_for_bias: bool = False
    categories: Optional[List[AttributeCategory]] = None
    bins: Optional[List[AttributeBin]] = None
    is_unique: bool = False
    is_positive_predicted_attribute: bool = False
    attribute_link: Optional[str] = None
    gt_class_link: Optional[str] = None
    implicit: bool = False

    def set(self, **kwargs):
        """Set one or many of the available properties of the ArthurAttribute class"""

        # do some basic validation on values to be set
        is_unique = kwargs.get("is_unique", self.categorical)
        categorical = kwargs.get("categorical", self.categorical)

        categories = [AttributeCategory(value=c) for c in kwargs.get("categories")] if "categories" in kwargs.keys() \
            else self.categories

        max_range = kwargs.get('max_range', self.max_range)
        min_range = kwargs.get('min_range', self.min_range)

        if categorical and not is_unique and categories is None:
            raise MissingParameterError("categories is required for non-unique categorical attributes")
        elif not categorical:
            if (min_range and max_range is None) or (min_range is None and max_range):
                raise MissingParameterError("Min and max range must both be set to either numerical values or none")
            elif min_range and max_range and min_range > max_range:
                raise UserValueError(f"Min range must be set to a value less then the max range, received the"
                                f" following (min, max) values: ({min_range}, {max_range})")
            elif categories:
                raise UserValueError("Categories are set for this attribute, they must be set to "
                                "None when categorical is False")

        # Check to make sure that the attributes which the user wants to set are valid
        for attribute_name, attribute_value in kwargs.items():
            if not ArthurAttribute.__annotations__.get(attribute_name):
                raise UserValueError(f"Attribute: '{attribute_name}' can not be set on an ArthurAttribute object")
            self.__setattr__(attribute_name, attribute_value)

            # format categories and bins appropriately
            if attribute_name == "categories":
                self.categories = categories
            if attribute_name == "bins":
                self.bins = [AttributeBin(kwargs.get("bins")[i], kwargs.get("bins")[i + 1])
                             for i in range(len(kwargs.get("bins")) - 1)] if "bins" in kwargs.keys() else self.bins

        return self

    def short_name(self):
        basename = f"{self.stage}[{self.position}]:{self.name}"
        label_suffix = f" ({self.label})" if self.label is not None else ""
        return basename + label_suffix

    def __str__(self):
        name = self.short_name()
        if self.categorical:
            name += f"\n\tCategorical: ["
            for cat in self.categories[:MAX_CATEGORIES_TO_SHOW]:
                name += f"{cat}, "
            name = name[:-2]
            if len(self.categories) > MAX_CATEGORIES_TO_SHOW:
                name += f"...and {len(self.categories) - MAX_CATEGORIES_TO_SHOW} more"
            name += "]"

        else:
            name += f"\n\tRange: [{self.min_range}, {self.max_range}]"

        if self.attribute_link:
            name += f"\n\tLink: {self.attribute_link}"

        if self.gt_class_link:
            name += f"\n\tLink: {self.gt_class_link}"
        return name



from datetime import timedelta, datetime
from typing import NamedTuple
from unittest import mock

import pytest
import pytz

from arthurai.client import auth
from arthurai.common.constants import API_PREFIX
from tests.fixtures.mocks import USER_TOKEN_JAN_1_2023_MIDNIGHT_EXPIRY

URL = "http://url"
LOGIN = "username"
PASSWORD = "password"
VERIFY_SSL = True


class AuthRefreshTestCase(NamedTuple):
    cur_time: datetime
    expected_wait_time: timedelta


CASES = [  # 2 hours to expiry: refresh in 115 minutes
         AuthRefreshTestCase(cur_time=datetime(2022, 12, 31, 22, 00, tzinfo=pytz.UTC),
                             expected_wait_time=timedelta(minutes=115)),
         # 30 min to expiry: refresh in 25 min
         AuthRefreshTestCase(cur_time=datetime(2022, 12, 31, 23, 30, tzinfo=pytz.UTC),
                             expected_wait_time=timedelta(minutes=25)),
         # 60 sec to expiry: refresh now
         AuthRefreshTestCase(cur_time=datetime(2022, 12, 31, 23, 59, tzinfo=pytz.UTC),
                             expected_wait_time=timedelta())
]


@mock.patch('arthurai.client.auth.user_login')
@mock.patch('arthurai.client.auth.datetime', wraps=datetime)
@pytest.mark.parametrize("case", CASES)
def test_refresh(mock_datetime, mock_user_login, case):
    # mock current time
    mock_datetime.now.return_value = case.cur_time
    # mock usre login
    mock_user_login.return_value = (USER_TOKEN_JAN_1_2023_MIDNIGHT_EXPIRY, {})

    # get value
    refresher = auth.AuthRefresher(url=URL, login=LOGIN, password=PASSWORD, verify_ssl=VERIFY_SSL)
    actual_header, actual_wait_time = refresher.refresh()

    # assert user login called with expected
    mock_user_login.assert_called_once_with(api_http_host=URL, api_prefix=API_PREFIX, login=LOGIN,
                                            password=PASSWORD, verify_ssl=VERIFY_SSL)
    # compare wait time to expected
    assert actual_wait_time == case.expected_wait_time
    # compare headers to expected
    assert actual_header == {"Authorization": USER_TOKEN_JAN_1_2023_MIDNIGHT_EXPIRY}
